# Copyright 2015 Tesora Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import enum
import hashlib
import os
from pathlib import Path
from requests.exceptions import ConnectionError

from oslo_log import log as logging
from swiftclient.client import ClientException

from trove.common import cfg
from trove.common import clients
from trove.common import exception
from trove.common.i18n import _
from trove.common import stream_codecs
from trove.common import timeutils
from trove.guestagent.common import operating_system
from trove.guestagent.common.operating_system import FileMode


LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class LogType(enum.Enum):
    """Represent the type of the log object."""

    # System logs.  These are always enabled.
    SYS = 1

    # User logs.  These can be enabled or disabled.
    USER = 2


class LogStatus(enum.Enum):
    """Represent the status of the log object."""

    # The log is disabled and potentially no data is being written to
    # the corresponding log file
    Disabled = 1

    # Logging is on, but no determination has been made about data availability
    Enabled = 2

    # Logging is on, but no log data is available to publish
    Unavailable = 3

    # Logging is on and data is available to be published
    Ready = 4

    # Logging is on and all data has been published
    Published = 5

    # Logging is on and some data has been published
    Partial = 6

    # Log file has been rotated, so next publish will discard log first
    Rotated = 7

    # Waiting for a datastore restart to begin logging
    Restart_Required = 8

    # Now that restart has completed, regular status can be reported again
    # This is an internal status
    Restart_Completed = 9


class GuestLog(object):

    MF_FILE_SUFFIX = '_metafile'
    MF_LABEL_LOG_NAME = 'log_name'
    MF_LABEL_LOG_TYPE = 'log_type'
    MF_LABEL_LOG_FILE = 'log_file'
    MF_LABEL_LOG_SIZE = 'log_size'
    MF_LABEL_LOG_HEADER = 'log_header_digest'

    def __init__(self, log_context, log_name, log_type, log_user, log_file,
                 log_exposed):
        self._context = log_context
        self._name = log_name
        self._type = log_type
        self._user = log_user
        self._file = log_file
        self._exposed = log_exposed
        self._size = None
        self._published_size = None
        self._header_digest = 'abc'
        self._published_header_digest = None
        self._status = None
        self._cached_context = None
        self._cached_swift_client = None
        self._enabled = log_type == LogType.SYS
        self._file_readable = False
        self._container_name = None
        self._codec = stream_codecs.JsonCodec()

        self._set_status(self._type == LogType.USER,
                         LogStatus.Disabled, LogStatus.Enabled)

        # The directory should already exist - make sure we have access to it
        log_dir = os.path.dirname(self._file)
        operating_system.chmod(
            log_dir, FileMode.ADD_GRP_RX_OTH_RX, as_root=True)

    @property
    def context(self):
        return self._context

    @context.setter
    def context(self, context):
        self._context = context

    @property
    def type(self):
        return self._type

    @property
    def swift_client(self):
        if not self._cached_swift_client or (
                self._cached_context != self.context):
            self._cached_swift_client = clients.swift_client(self.context)
            self._cached_context = self.context
        return self._cached_swift_client

    @property
    def exposed(self):
        return self._exposed or self.context.is_admin

    @property
    def enabled(self):
        return self._enabled

    @enabled.setter
    def enabled(self, enabled):
        self._enabled = enabled

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, status):
        # Keep the status in Restart_Required until we're set
        # to Restart_Completed
        if (self.status != LogStatus.Restart_Required or
                (self.status == LogStatus.Restart_Required and
                 status == LogStatus.Restart_Completed)):
            self._status = status
            LOG.debug("Log status for '%(name)s' set to %(status)s",
                      {'name': self._name, 'status': status})
        else:
            LOG.debug("Log status for '%(name)s' *not* set to %(status)s "
                      "(currently %(current_status)s)",
                      {'name': self._name, 'status': status,
                       'current_status': self.status})

    def get_container_name(self, force=False):
        if not self._container_name or force:
            container_name = CONF.guest_log_container_name
            try:
                self.swift_client.get_container(container_name, prefix='dummy')
            except ClientException as ex:
                if ex.http_status == 404:
                    LOG.debug("Container '%s' not found; creating now",
                              container_name)
                    self.swift_client.put_container(
                        container_name, headers=self._get_headers())
                else:
                    LOG.exception("Could not retrieve container '%s'",
                                  container_name)
                    raise
            self._container_name = container_name
        return self._container_name

    def _set_status(self, use_first, first_status, second_status):
        if use_first:
            self.status = first_status
        else:
            self.status = second_status

    def show(self):
        if self.exposed:
            self._refresh_details()
            container_name = 'None'
            prefix = 'None'
            if self._published_size:
                container_name = self.get_container_name()
                prefix = self._object_prefix()
            pending = self._size - self._published_size
            if self.status == LogStatus.Rotated:
                pending = self._size
            return {
                'name': self._name,
                'type': self._type.name,
                'status': self.status.name.replace('_', ' '),
                'published': self._published_size,
                'pending': pending,
                'container': container_name,
                'prefix': prefix,
                'metafile': self._metafile_name()
            }
        else:
            raise exception.LogAccessForbidden(action='show', log=self._name)

    def _refresh_details(self):
        if self._published_size is None:
            # Initializing, so get all the values
            try:
                meta_details = self._get_meta_details()
                self._published_size = int(
                    meta_details[self.MF_LABEL_LOG_SIZE])
                self._published_header_digest = (
                    meta_details[self.MF_LABEL_LOG_HEADER])
            except ClientException as ex:
                if ex.http_status == 404:
                    LOG.debug("No published metadata found for log '%s'",
                              self._name)
                    self._published_size = 0
                else:
                    LOG.exception("Could not get meta details for log '%s'",
                                  self._name)
                    raise
            except ConnectionError as e:
                # A bad endpoint will cause a ConnectionError
                # This exception contains another exception that we want
                exc = e.args[0]
                raise exc

        self._update_details()
        LOG.debug("Log size for '%(name)s' set to %(size)d "
                  "(published %(published)d)",
                  {'name': self._name, 'size': self._size,
                   'published': self._published_size})

    def _update_details(self):
        if operating_system.exists(self._file, as_root=True):
            file_path = Path(self._file)

            # Make sure guest agent can read the log file.
            if not os.access(self._file, os.R_OK):
                operating_system.chmod(self._file, FileMode.ADD_ALL_R,
                                       as_root=True)
                operating_system.chmod(str(file_path.parent),
                                       FileMode.ADD_GRP_RX_OTH_RX,
                                       as_root=True)

            self._size = file_path.stat().st_size
            self._update_log_header_digest(self._file)

            if self.status != LogStatus.Disabled:
                if self._log_rotated():
                    self.status = LogStatus.Rotated
                # See if we have stuff to publish
                elif self._size > self._published_size:
                    self._set_status(self._published_size,
                                     LogStatus.Partial, LogStatus.Ready)
                # We've published everything so far
                elif self._size == self._published_size:
                    self._set_status(self._published_size,
                                     LogStatus.Published, LogStatus.Enabled)
                # We've already handled this case (log rotated) so what gives?
                else:
                    raise Exception(_("Bug in _log_rotated ?"))
        else:
            LOG.warning(f"File {self._file} does not exist")
            self._published_size = 0
            self._size = 0

    def _log_rotated(self):
        """If the file is smaller than the last reported size
        or the first line hash is different, we can probably assume
        the file changed under our nose.
        """
        if (self._published_size > 0 and
                (self._size < self._published_size or
                 self._published_header_digest != self._header_digest)):
            return True

    def _update_log_header_digest(self, log_file):
        with open(log_file, 'rb') as log:
            self._header_digest = hashlib.md5(log.readline()).hexdigest()

    def _get_headers(self):
        return {'X-Delete-After': str(CONF.guest_log_expiry)}

    def publish_log(self):
        if self.exposed:
            if self._log_rotated():
                LOG.debug("Log file rotation detected for '%s' - "
                          "discarding old log", self._name)
                self._delete_log_components()
            if operating_system.exists(self._file, as_root=True):
                self._publish_to_container(self._file)
            else:
                raise RuntimeError(_(
                    "Cannot publish log file '%s' as it does not exist.") %
                    self._file)
            return self.show()
        else:
            raise exception.LogAccessForbidden(
                action='publish', log=self._name)

    def discard_log(self):
        if self.exposed:
            self._delete_log_components()
            return self.show()
        else:
            raise exception.LogAccessForbidden(
                action='discard', log=self._name)

    def _delete_log_components(self):
        container_name = self.get_container_name(force=True)
        prefix = self._object_prefix()
        swift_files = [swift_file['name']
                       for swift_file in self.swift_client.get_container(
                       container_name, prefix=prefix)[1]]
        swift_files.append(self._metafile_name())
        for swift_file in swift_files:
            self.swift_client.delete_object(container_name, swift_file)
        self._published_size = 0

    def _publish_to_container(self, log_filename):
        log_component, log_lines = '', 0
        chunk_size = CONF.guest_log_limit
        container_name = self.get_container_name(force=True)

        def _read_chunk(f):
            while True:
                current_chunk = f.read(chunk_size)
                if not current_chunk:
                    break
                yield current_chunk

        def _write_log_component():
            object_headers.update({'x-object-meta-lines': str(log_lines)})
            component_name = '%s%s' % (self._object_prefix(),
                                       self._object_name())
            self.swift_client.put_object(container_name,
                                         component_name, log_component,
                                         headers=object_headers)
            self._published_size = (
                self._published_size + len(log_component))
            self._published_header_digest = self._header_digest

        self._refresh_details()
        self._put_meta_details()
        object_headers = self._get_headers()
        with open(log_filename, 'r') as log:
            LOG.debug("seeking to %s", self._published_size)
            log.seek(self._published_size)
            for chunk in _read_chunk(log):
                for log_line in chunk.splitlines():
                    if len(log_component) + len(log_line) > chunk_size:
                        _write_log_component()
                        log_component, log_lines = '', 0
                    log_component = log_component + log_line + '\n'
                    log_lines += 1
        if log_lines > 0:
            _write_log_component()
        self._put_meta_details()

    def _put_meta_details(self):
        metafile_name = self._metafile_name()
        metafile_details = {
            self.MF_LABEL_LOG_NAME: self._name,
            self.MF_LABEL_LOG_TYPE: self._type.name,
            self.MF_LABEL_LOG_FILE: self._file,
            self.MF_LABEL_LOG_SIZE: self._published_size,
            self.MF_LABEL_LOG_HEADER: self._header_digest,
        }
        container_name = self.get_container_name()
        self.swift_client.put_object(container_name, metafile_name,
                                     self._codec.serialize(metafile_details),
                                     headers=self._get_headers())
        LOG.debug("_put_meta_details has published log size as %s",
                  self._published_size)

    def _metafile_name(self):
        return self._object_prefix().rstrip('/') + '_metafile'

    def _object_prefix(self):
        return '%(instance_id)s/%(datastore)s-%(log)s/' % {
            'instance_id': CONF.guest_id,
            'datastore': CONF.datastore_manager,
            'log': self._name}

    def _object_name(self):
        return 'log-%s' % str(timeutils.utcnow()).replace(' ', 'T')

    def _get_meta_details(self):
        LOG.debug("Getting meta details for '%s'", self._name)
        metafile_name = self._metafile_name()
        container_name = self.get_container_name()
        headers, metafile_details = self.swift_client.get_object(
            container_name, metafile_name)
        LOG.debug("Found meta details for '%s'", self._name)
        return self._codec.deserialize(metafile_details)



# Copyright [2013] Hewlett-Packard Development Company, L.P.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sqlalchemy.schema import Column
from sqlalchemy.schema import MetaData

from trove.db.sqlalchemy.migrate_repo.schema import Boolean
from trove.db.sqlalchemy.migrate_repo.schema import create_tables
from trove.db.sqlalchemy.migrate_repo.schema import DateTime
from trove.db.sqlalchemy.migrate_repo.schema import Float
from trove.db.sqlalchemy.migrate_repo.schema import String
from trove.db.sqlalchemy.migrate_repo.schema import Table

meta = MetaData()

backups = Table('backups', meta,
                Column('id', String(36), primary_key=True, nullable=False),
                Column('name', String(255), nullable=False),
                Column('description', String(512)),
                Column('location', String(1024)),
                Column('backup_type', String(32)),
                Column('size', Float()),
                Column('tenant_id', String(36)),
                Column('state', String(32), nullable=False),
                Column('instance_id', String(36)),
                Column('checksum', String(32)),
                Column('backup_timestamp', DateTime()),
                Column('deleted', Boolean()),
                Column('created', DateTime()),
                Column('updated', DateTime()),
                Column('deleted_at', DateTime()))


def upgrade(migrate_engine):
    meta.bind = migrate_engine
    create_tables([backups, ])



# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

import trove.common.apischema as apischema
from trove.common.auth import admin_context
from trove.common import wsgi
from trove.extensions.mgmt.upgrade.models import UpgradeMessageSender


LOG = logging.getLogger(__name__)


class UpgradeController(wsgi.Controller):
    """
    Controller for guest agent upgrade
    """
    schemas = apischema.upgrade

    @admin_context
    def create(self, req, body, tenant_id, instance_id):
        LOG.info("Sending upgrade notifications\nreq : '%(req)s'\n"
                 "Admin tenant_id: %(tenant_id)s",
                 {"tenant_id": tenant_id, "req": req})

        context = req.environ.get(wsgi.CONTEXT_KEY)
        upgrade = body['upgrade']

        instance_version = upgrade.get('instance_version')
        location = upgrade.get('location')
        metadata = upgrade.get('metadata')

        send = UpgradeMessageSender.create(
            context, instance_id, instance_version, location, metadata)

        send()
        return wsgi.Result(None, 202)



# Copyright [2015] Hewlett-Packard Development Company, L.P.
# Copyright 2016 Tesora Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from eventlet.timeout import Timeout
from oslo_log import log as logging

from trove.common import cfg
from trove.common.clients import create_nova_client
from trove.common.exception import PollTimeOut
from trove.common.exception import TroveError
from trove.common.i18n import _
from trove.common.strategies.cluster import base as cluster_base
from trove.common.template import ClusterConfigTemplate
from trove.common import utils
from trove.extensions.common import models as ext_models
from trove.instance.models import DBInstance
from trove.instance.models import Instance
from trove.instance import tasks as inst_tasks
from trove.taskmanager import api as task_api
import trove.taskmanager.models as task_models


LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class GaleraCommonTaskManagerStrategy(cluster_base.BaseTaskManagerStrategy):

    @property
    def task_manager_api_class(self):
        return task_api.API

    @property
    def task_manager_cluster_tasks_class(self):
        return GaleraCommonClusterTasks


class GaleraCommonClusterTasks(task_models.ClusterTasks):

    CLUSTER_REPLICATION_USER = "clusterrepuser"

    def _render_cluster_config(self, context, instance, cluster_ips,
                               cluster_name, replication_user):
        client = create_nova_client(context)
        flavor = client.flavors.get(instance.flavor_id)
        instance_ip = self.get_ip(instance)
        config = ClusterConfigTemplate(
            self.datastore_version, flavor, instance.id)
        replication_user_pass = "%(name)s:%(password)s" % replication_user
        config_rendered = config.render(
            replication_user_pass=replication_user_pass,
            cluster_ips=cluster_ips,
            cluster_name=cluster_name,
            instance_ip=instance_ip,
            instance_name=instance.name,
        )
        return config_rendered

    def create_cluster(self, context, cluster_id):
        LOG.debug("Begin create_cluster for id: %s.", cluster_id)

        def _create_cluster():
            # Fetch instances by cluster_id against instances table.
            db_instances = DBInstance.find_all(cluster_id=cluster_id).all()
            instance_ids = [db_instance.id for db_instance in db_instances]

            LOG.debug("Waiting for instances to get to cluster-ready status.")
            # Wait for cluster members to get to cluster-ready status.
            if not self._all_instances_ready(instance_ids, cluster_id):
                raise TroveError(_("Instances in cluster did not report "
                                   "ACTIVE"))

            LOG.debug("All members ready, proceeding for cluster setup.")
            instances = [Instance.load(context, instance_id) for instance_id
                         in instance_ids]

            cluster_ips = [self.get_ip(instance) for instance in instances]
            instance_guests = []

            # Create replication user and password for synchronizing the
            # galera cluster
            replication_user = {
                "name": self.CLUSTER_REPLICATION_USER,
                "password": utils.generate_random_password(),
            }

            # Galera cluster name must be unique and be shorter than a full
            # uuid string so we remove the hyphens and chop it off. It was
            # recommended to be 16 chars or less.
            # (this is not currently documented on Galera docs)
            cluster_name = utils.generate_uuid().replace("-", "")[:16]

            LOG.debug("Configuring cluster configuration.")
            try:
                # Set the admin password for all the instances because the
                # password in the my.cnf will be wrong after the joiner
                # instances syncs with the donor instance.
                admin_password = str(utils.generate_random_password())

                bootstrap = True
                for instance in instances:
                    guest = self.get_guest(instance)
                    instance_guests.append(guest)
                    guest.reset_admin_password(admin_password)
                    # render the conf.d/cluster.cnf configuration
                    cluster_configuration = self._render_cluster_config(
                        context,
                        instance,
                        ",".join(cluster_ips),
                        cluster_name,
                        replication_user)

                    # push the cluster config and bootstrap the first instance
                    guest.install_cluster(replication_user,
                                          cluster_configuration,
                                          bootstrap)
                    bootstrap = False

                LOG.debug("Finalizing cluster configuration.")
                for guest in instance_guests:
                    guest.cluster_complete()
            except Exception:
                LOG.exception("Error creating cluster.")
                self.update_statuses_on_failure(cluster_id)

        timeout = Timeout(CONF.cluster_usage_timeout)
        try:
            _create_cluster()
            self.reset_task()
        except Timeout as t:
            if t is not timeout:
                raise  # not my timeout
            LOG.exception("Timeout for building cluster.")
            self.update_statuses_on_failure(cluster_id)
        except TroveError:
            LOG.exception("Error creating cluster %s.", cluster_id)
            self.update_statuses_on_failure(cluster_id)
        finally:
            timeout.cancel()

        LOG.debug("End create_cluster for id: %s.", cluster_id)

    def _check_cluster_for_root(self, context, existing_instances,
                                new_instances):
        """Check for existing instances root enabled"""
        for instance in existing_instances:
            if ext_models.Root.load(context, instance.id):
                for new_instance in new_instances:
                    ext_models.RootHistory.create(context, new_instance.id)
                return

    def grow_cluster(self, context, cluster_id, new_instance_ids):
        LOG.debug("Begin Galera grow_cluster for id: %s.", cluster_id)

        def _grow_cluster():

            db_instances = DBInstance.find_all(
                cluster_id=cluster_id, deleted=False).all()
            existing_instances = [Instance.load(context, db_inst.id)
                                  for db_inst in db_instances
                                  if db_inst.id not in new_instance_ids]
            if not existing_instances:
                raise TroveError(_("Unable to determine existing cluster "
                                   "member(s)"))

            # get list of ips of existing cluster members
            existing_cluster_ips = [self.get_ip(instance) for instance in
                                    existing_instances]
            existing_instance_guests = [self.get_guest(instance)
                                        for instance in existing_instances]

            # get the cluster context to setup new members
            cluster_context = existing_instance_guests[0].get_cluster_context()

            # Wait for cluster members to get to cluster-ready status.
            if not self._all_instances_ready(new_instance_ids, cluster_id):
                raise TroveError(_("Instances in cluster did not report "
                                   "ACTIVE"))

            LOG.debug("All members ready, proceeding for cluster setup.")

            # Get the new instances to join the cluster
            new_instances = [Instance.load(context, instance_id)
                             for instance_id in new_instance_ids]
            new_cluster_ips = [self.get_ip(instance) for instance in
                               new_instances]
            for instance in new_instances:
                guest = self.get_guest(instance)

                guest.reset_admin_password(cluster_context['admin_password'])

                # render the conf.d/cluster.cnf configuration
                cluster_configuration = self._render_cluster_config(
                    context,
                    instance,
                    ",".join(existing_cluster_ips),
                    cluster_context['cluster_name'],
                    cluster_context['replication_user'])

                # push the cluster config and bootstrap the first instance
                bootstrap = False
                guest.install_cluster(cluster_context['replication_user'],
                                      cluster_configuration,
                                      bootstrap)

            self._check_cluster_for_root(context,
                                         existing_instances,
                                         new_instances)

            # apply the new config to all instances
            for instance in existing_instances + new_instances:
                guest = self.get_guest(instance)
                # render the conf.d/cluster.cnf configuration
                cluster_configuration = self._render_cluster_config(
                    context,
                    instance,
                    ",".join(existing_cluster_ips + new_cluster_ips),
                    cluster_context['cluster_name'],
                    cluster_context['replication_user'])
                guest.write_cluster_configuration_overrides(
                    cluster_configuration)

            for instance in new_instances:
                guest = self.get_guest(instance)
                guest.cluster_complete()

        timeout = Timeout(CONF.cluster_usage_timeout)
        try:
            _grow_cluster()
            self.reset_task()
        except Timeout as t:
            if t is not timeout:
                raise  # not my timeout
            LOG.exception("Timeout for growing cluster.")
            self.update_statuses_on_failure(
                cluster_id, status=inst_tasks.InstanceTasks.GROWING_ERROR)
        except Exception:
            LOG.exception("Error growing cluster %s.", cluster_id)
            self.update_statuses_on_failure(
                cluster_id, status=inst_tasks.InstanceTasks.GROWING_ERROR)
        finally:
            timeout.cancel()

        LOG.debug("End grow_cluster for id: %s.", cluster_id)

    def shrink_cluster(self, context, cluster_id, removal_instance_ids):
        LOG.debug("Begin Galera shrink_cluster for id: %s.", cluster_id)

        def _shrink_cluster():
            removal_instances = [Instance.load(context, instance_id)
                                 for instance_id in removal_instance_ids]
            for instance in removal_instances:
                Instance.delete(instance)

            # wait for instances to be deleted
            def all_instances_marked_deleted():
                non_deleted_instances = DBInstance.find_all(
                    cluster_id=cluster_id, deleted=False).all()
                non_deleted_ids = [db_instance.id for db_instance
                                   in non_deleted_instances]
                return not bool(
                    set(removal_instance_ids).intersection(
                        set(non_deleted_ids))
                )
            try:
                LOG.info("Deleting instances (%s)", removal_instance_ids)
                utils.poll_until(all_instances_marked_deleted,
                                 sleep_time=2,
                                 time_out=CONF.cluster_delete_time_out)
            except PollTimeOut:
                LOG.error("timeout for instances to be marked as deleted.")
                return

            db_instances = DBInstance.find_all(
                cluster_id=cluster_id, deleted=False).all()
            leftover_instances = [Instance.load(context, db_inst.id)
                                  for db_inst in db_instances
                                  if db_inst.id not in removal_instance_ids]
            leftover_cluster_ips = [self.get_ip(instance) for instance in
                                    leftover_instances]

            # Get config changes for left over instances
            rnd_cluster_guest = self.get_guest(leftover_instances[0])
            cluster_context = rnd_cluster_guest.get_cluster_context()

            # apply the new config to all leftover instances
            for instance in leftover_instances:
                guest = self.get_guest(instance)
                # render the conf.d/cluster.cnf configuration
                cluster_configuration = self._render_cluster_config(
                    context,
                    instance,
                    ",".join(leftover_cluster_ips),
                    cluster_context['cluster_name'],
                    cluster_context['replication_user'])
                guest.write_cluster_configuration_overrides(
                    cluster_configuration)

        timeout = Timeout(CONF.cluster_usage_timeout)
        try:
            _shrink_cluster()
            self.reset_task()
        except Timeout as t:
            if t is not timeout:
                raise  # not my timeout
            LOG.exception("Timeout for shrinking cluster.")
            self.update_statuses_on_failure(
                cluster_id, status=inst_tasks.InstanceTasks.SHRINKING_ERROR)
        except Exception:
            LOG.exception("Error shrinking cluster %s.", cluster_id)
            self.update_statuses_on_failure(
                cluster_id, status=inst_tasks.InstanceTasks.SHRINKING_ERROR)
        finally:
            timeout.cancel()

        LOG.debug("End shrink_cluster for id: %s.", cluster_id)

    def restart_cluster(self, context, cluster_id):
        self.rolling_restart_cluster(context, cluster_id)

    def upgrade_cluster(self, context, cluster_id, datastore_version):
        self.rolling_upgrade_cluster(context, cluster_id, datastore_version)



# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from proboscis import SkipTest

from trove.tests.scenario.runners.test_runners import TestRunner
from troveclient.compat import exceptions


class NegativeClusterActionsRunner(TestRunner):

    def __init__(self):
        super(NegativeClusterActionsRunner, self).__init__()

    def run_create_constrained_size_cluster(self, min_nodes=2, max_nodes=None,
                                            expected_http_code=400):
        self.assert_create_constrained_size_cluster('negative_cluster',
                                                    min_nodes, max_nodes,
                                                    expected_http_code)

    def assert_create_constrained_size_cluster(self, cluster_name,
                                               min_nodes, max_nodes,
                                               expected_http_code):
        # Create a cluster with less than 'min_nodes'.
        if min_nodes:
            instances_def = [self.build_flavor()] * (min_nodes - 1)
            self._assert_cluster_create_raises(cluster_name, instances_def,
                                               expected_http_code)

        # Create a cluster with mare than 'max_nodes'.
        if max_nodes:
            instances_def = [self.build_flavor()] * (max_nodes + 1)
            self._assert_cluster_create_raises(cluster_name, instances_def,
                                               expected_http_code)

    def run_create_heterogeneous_cluster(self, expected_http_code=400):
        # Create a cluster with different node flavors.
        instances_def = [self.build_flavor(flavor_id=2, volume_size=1),
                         self.build_flavor(flavor_id=3, volume_size=1)]
        self._assert_cluster_create_raises('heterocluster',
                                           instances_def, expected_http_code)

        # Create a cluster with different volume sizes.
        instances_def = [self.build_flavor(flavor_id=2, volume_size=1),
                         self.build_flavor(flavor_id=2, volume_size=2)]
        self._assert_cluster_create_raises('heterocluster',
                                           instances_def, expected_http_code)

    def _assert_cluster_create_raises(self, cluster_name, instances_def,
                                      expected_http_code):
        client = self.auth_client
        self.assert_raises(exceptions.BadRequest, expected_http_code,
                           client, client.clusters.create,
                           cluster_name,
                           self.instance_info.dbaas_datastore,
                           self.instance_info.dbaas_datastore_version,
                           instances=instances_def)


class MongodbNegativeClusterActionsRunner(NegativeClusterActionsRunner):

    def run_create_constrained_size_cluster(self):
        super(MongodbNegativeClusterActionsRunner,
              self).run_create_constrained_size_cluster(min_nodes=3,
                                                        max_nodes=3)


class CassandraNegativeClusterActionsRunner(NegativeClusterActionsRunner):

    def run_create_constrained_size_cluster(self):
        raise SkipTest("No constraints apply to the number of cluster nodes.")

    def run_create_heterogeneous_cluster(self):
        raise SkipTest("No constraints apply to the size of cluster nodes.")


class RedisNegativeClusterActionsRunner(NegativeClusterActionsRunner):

    def run_create_constrained_size_cluster(self):
        raise SkipTest("No constraints apply to the number of cluster nodes.")

    def run_create_heterogeneous_cluster(self):
        raise SkipTest("No constraints apply to the size of cluster nodes.")


class PxcNegativeClusterActionsRunner(NegativeClusterActionsRunner):

    def run_create_constrained_size_cluster(self):
        raise SkipTest("No constraints apply to the number of cluster nodes.")



# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from proboscis import test

from trove.tests.scenario import groups
from trove.tests.scenario.groups.test_group import TestGroup
from trove.tests.scenario.runners import test_runners


GROUP = "scenario.database_actions_group"


class DatabaseActionsRunnerFactory(test_runners.RunnerFactory):

    _runner_ns = 'database_actions_runners'
    _runner_cls = 'DatabaseActionsRunner'


class InstanceCreateRunnerFactory(test_runners.RunnerFactory):

    _runner_ns = 'instance_create_runners'
    _runner_cls = 'InstanceCreateRunner'


@test(depends_on_groups=[groups.CFGGRP_DELETE],
      groups=[GROUP, groups.DB_ACTION_CREATE])
class DatabaseActionsCreateGroup(TestGroup):
    """Test Database Actions Create functionality."""

    def __init__(self):
        super(DatabaseActionsCreateGroup, self).__init__(
            DatabaseActionsRunnerFactory.instance())

    @test
    def create_databases(self):
        """Create databases on an existing instance."""
        self.test_runner.run_databases_create()

    @test(depends_on=[create_databases])
    def list_databases(self):
        """List the created databases."""
        self.test_runner.run_databases_list()

    @test(depends_on=[create_databases],
          runs_after=[list_databases])
    def create_database_with_no_attributes(self):
        """Ensure creating a database with blank specification fails."""
        self.test_runner.run_database_create_with_no_attributes()

    @test(depends_on=[create_databases],
          runs_after=[create_database_with_no_attributes])
    def create_database_with_blank_name(self):
        """Ensure creating a database with blank name fails."""
        self.test_runner.run_database_create_with_blank_name()

    @test(depends_on=[create_databases],
          runs_after=[create_database_with_blank_name])
    def create_existing_database(self):
        """Ensure creating an existing database fails."""
        self.test_runner.run_existing_database_create()


@test(depends_on_classes=[DatabaseActionsCreateGroup],
      groups=[GROUP, groups.DB_ACTION_DELETE])
class DatabaseActionsDeleteGroup(TestGroup):
    """Test Database Actions Delete functionality."""

    def __init__(self):
        super(DatabaseActionsDeleteGroup, self).__init__(
            DatabaseActionsRunnerFactory.instance())

    @test
    def delete_database(self):
        """Delete the created databases."""
        self.test_runner.run_database_delete()

    @test(runs_after=[delete_database])
    def delete_nonexisting_database(self):
        """Delete non-existing databases."""
        self.test_runner.run_nonexisting_database_delete()

    @test(runs_after=[delete_nonexisting_database])
    def create_system_database(self):
        """Ensure creating a system database fails."""
        self.test_runner.run_system_database_create()

    @test(runs_after=[create_system_database])
    def delete_system_database(self):
        """Ensure deleting a system database fails."""
        self.test_runner.run_system_database_delete()


@test(depends_on_classes=[DatabaseActionsDeleteGroup],
      groups=[GROUP, groups.DB_ACTION_INST, groups.DB_ACTION_INST_CREATE])
class DatabaseActionsInstCreateGroup(TestGroup):
    """Test Database Actions Instance Create functionality."""

    def __init__(self):
        super(DatabaseActionsInstCreateGroup, self).__init__(
            DatabaseActionsRunnerFactory.instance())
        self.instance_create_runner = InstanceCreateRunnerFactory.instance()

    @test
    def create_initialized_instance(self):
        """Create an instance with initial databases."""
        self.instance_create_runner.run_initialized_instance_create(
            with_dbs=True, with_users=False, configuration_id=None,
            name_suffix='_db')


@test(depends_on_classes=[DatabaseActionsInstCreateGroup],
      groups=[GROUP, groups.DB_ACTION_INST, groups.DB_ACTION_INST_CREATE_WAIT])
class DatabaseActionsInstCreateWaitGroup(TestGroup):
    """Wait for Database Actions Instance Create to complete."""

    def __init__(self):
        super(DatabaseActionsInstCreateWaitGroup, self).__init__(
            DatabaseActionsRunnerFactory.instance())
        self.instance_create_runner = InstanceCreateRunnerFactory.instance()

    @test
    def wait_for_instances(self):
        """Waiting for database instance to become active."""
        self.instance_create_runner.run_wait_for_init_instance()

    @test(depends_on=[wait_for_instances])
    def add_initialized_instance_data(self):
        """Add data to the database instance."""
        self.instance_create_runner.run_add_initialized_instance_data()

    @test(runs_after=[add_initialized_instance_data])
    def validate_initialized_instance(self):
        """Validate the database instance data and properties."""
        self.instance_create_runner.run_validate_initialized_instance()


@test(depends_on_classes=[DatabaseActionsInstCreateWaitGroup],
      groups=[GROUP, groups.DB_ACTION_INST, groups.DB_ACTION_INST_DELETE])
class DatabaseActionsInstDeleteGroup(TestGroup):
    """Test Database Actions Instance Delete functionality."""

    def __init__(self):
        super(DatabaseActionsInstDeleteGroup, self).__init__(
            DatabaseActionsRunnerFactory.instance())
        self.instance_create_runner = InstanceCreateRunnerFactory.instance()

    @test
    def delete_initialized_instance(self):
        """Delete the database instance."""
        self.instance_create_runner.run_initialized_instance_delete()


@test(depends_on_classes=[DatabaseActionsInstDeleteGroup],
      groups=[GROUP, groups.DB_ACTION_INST, groups.DB_ACTION_INST_DELETE_WAIT])
class DatabaseActionsInstDeleteWaitGroup(TestGroup):
    """Wait for Database Actions Instance Delete to complete."""

    def __init__(self):
        super(DatabaseActionsInstDeleteWaitGroup, self).__init__(
            DatabaseActionsRunnerFactory.instance())
        self.instance_create_runner = InstanceCreateRunnerFactory.instance()

    @test
    def wait_for_delete_initialized_instance(self):
        """Wait for the database instance to delete."""
        self.instance_create_runner.run_wait_for_init_delete()



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy import MetaData
from sqlalchemy import orm
from sqlalchemy.orm import exc as orm_exc
from sqlalchemy import Table


def map(engine, models):
    meta = MetaData()
    meta.bind = engine
    if mapping_exists(models['instances']):
        return

    orm.mapper(models['instances'], Table('instances', meta, autoload=True))
    orm.mapper(models['instance_faults'],
               Table('instance_faults', meta, autoload=True))
    orm.mapper(models['root_enabled_history'],
               Table('root_enabled_history', meta, autoload=True))
    orm.mapper(models['datastores'],
               Table('datastores', meta, autoload=True))
    orm.mapper(models['datastore_versions'],
               Table('datastore_versions', meta, autoload=True))
    orm.mapper(models['datastore_version_metadata'],
               Table('datastore_version_metadata', meta, autoload=True))
    orm.mapper(models['capabilities'],
               Table('capabilities', meta, autoload=True))
    orm.mapper(models['capability_overrides'],
               Table('capability_overrides', meta, autoload=True))
    orm.mapper(models['service_statuses'],
               Table('service_statuses', meta, autoload=True))
    orm.mapper(models['dns_records'],
               Table('dns_records', meta, autoload=True))
    orm.mapper(models['agent_heartbeats'],
               Table('agent_heartbeats', meta, autoload=True))
    orm.mapper(models['quotas'],
               Table('quotas', meta, autoload=True))
    orm.mapper(models['quota_usages'],
               Table('quota_usages', meta, autoload=True))
    orm.mapper(models['reservations'],
               Table('reservations', meta, autoload=True))
    orm.mapper(models['backups'],
               Table('backups', meta, autoload=True))
    orm.mapper(models['backup_strategy'],
               Table('backup_strategy', meta, autoload=True))
    orm.mapper(models['security_groups'],
               Table('security_groups', meta, autoload=True))
    orm.mapper(models['security_group_rules'],
               Table('security_group_rules', meta, autoload=True))
    orm.mapper(models['security_group_instance_associations'],
               Table('security_group_instance_associations', meta,
                     autoload=True))
    orm.mapper(models['configurations'],
               Table('configurations', meta, autoload=True))
    orm.mapper(models['configuration_parameters'],
               Table('configuration_parameters', meta, autoload=True))
    orm.mapper(models['conductor_lastseen'],
               Table('conductor_lastseen', meta, autoload=True))
    orm.mapper(models['clusters'],
               Table('clusters', meta, autoload=True))
    orm.mapper(models['datastore_configuration_parameters'],
               Table('datastore_configuration_parameters', meta,
                     autoload=True))
    orm.mapper(models['modules'],
               Table('modules', meta, autoload=True))
    orm.mapper(models['instance_modules'],
               Table('instance_modules', meta, autoload=True))


def mapping_exists(model):
    try:
        orm.class_mapper(model)
        return True
    except orm_exc.UnmappedClassError:
        return False



# Copyright 2020 Catalyst Cloud
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from migrate.changeset.constraint import UniqueConstraint
from sqlalchemy import text
from sqlalchemy.schema import Column
from sqlalchemy.schema import MetaData
from sqlalchemy.sql.expression import select
from sqlalchemy.sql.expression import update

from trove.db.sqlalchemy import utils as db_utils
from trove.db.sqlalchemy.migrate_repo.schema import String
from trove.db.sqlalchemy.migrate_repo.schema import Table


def upgrade(migrate_engine):
    meta = MetaData()
    meta.bind = migrate_engine

    ds_table = Table('datastores', meta, autoload=True)
    ds_version_table = Table('datastore_versions', meta, autoload=True)
    ds_version_table.create_column(
        Column('version', String(255), nullable=True))

    ds_versions = select(
        columns=[text("id"), text("name")],
        from_obj=ds_version_table
    ).execute()

    # Use 'name' value as init 'version' value
    for version in ds_versions:
        update(
            table=ds_version_table,
            whereclause=text("id='%s'" % version.id),
            values=dict(version=version.name)
        ).execute()

    # Change unique constraint, need to drop the foreign key first and add back
    # later
    constraint_names = db_utils.get_foreign_key_constraint_names(
        engine=migrate_engine,
        table='datastore_versions',
        columns=['datastore_id'],
        ref_table='datastores',
        ref_columns=['id'])
    db_utils.drop_foreign_key_constraints(
        constraint_names=constraint_names,
        columns=[ds_version_table.c.datastore_id],
        ref_columns=[ds_table.c.id])

    UniqueConstraint('datastore_id', 'name', name='ds_versions',
                     table=ds_version_table).drop()
    UniqueConstraint('datastore_id', 'name', 'version', name='ds_versions',
                     table=ds_version_table).create()

    db_utils.create_foreign_key_constraints(
        constraint_names=constraint_names,
        columns=[ds_version_table.c.datastore_id],
        ref_columns=[ds_table.c.id])



# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from oslo_service import loopingcall


def build_polling_task(retriever, condition=lambda value: value,
                       sleep_time=1, time_out=0, initial_delay=0):
    """Run a function in a loop with backoff on error.

    The condition function runs based on the retriever function result.
    """

    def poll_and_check():
        obj = retriever()
        if condition(obj):
            raise loopingcall.LoopingCallDone(retvalue=obj)

    call = loopingcall.BackOffLoopingCall(f=poll_and_check)
    return call.start(initial_delay=initial_delay,
                      starting_interval=sleep_time,
                      max_interval=30, timeout=time_out)


def poll_until(retriever, condition=lambda value: value,
               sleep_time=3, time_out=0, initial_delay=0):
    """Retrieves object until it passes condition, then returns it.

    If time_out_limit is passed in, PollTimeOut will be raised once that
    amount of time is eclipsed.

    """
    task = build_polling_task(retriever, condition=condition,
                              sleep_time=sleep_time, time_out=time_out,
                              initial_delay=initial_delay)
    return task.wait()



# Copyright (c) 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import os
import shlex
from tempfile import NamedTemporaryFile
import traceback

from oslo_log import log as logging

from trove.common import cfg
from trove.common import exception
from trove.common.i18n import _
from trove.common import utils
from trove.guestagent.common import operating_system

TMP_MOUNT_POINT = "/mnt/volume"

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


# We removed all translation for messages destinated to log file.
# However we cannot use _(xxx) instead of _("xxxx") because of the
# H701 pep8 checking, so we have to pass different message format
# string and format content here.
def log_and_raise(log_fmt, exc_fmt, fmt_content=None):
    if fmt_content is not None:
        LOG.exception(log_fmt, fmt_content)
        raise_msg = exc_fmt % fmt_content
    else:
        # if fmt_content is not provided, log_fmt and
        # exc_fmt are just plain string messages
        LOG.exception(log_fmt)
        raise_msg = exc_fmt
    raise_msg += _("\nExc: %s") % traceback.format_exc()
    raise exception.GuestError(original_message=raise_msg)


class FSBase(object, metaclass=abc.ABCMeta):

    def __init__(self, fstype, format_options):
        self.fstype = fstype
        self.format_options = format_options

    @abc.abstractmethod
    def format(self, device_path, timeout):
        """
        Format device
        """

    @abc.abstractmethod
    def check_format(self, device_path):
        """
        Check if device is formatted
        """

    @abc.abstractmethod
    def resize(self, device_path, online=False):
        """
        Resize the filesystem on device
        """


class FSExt(FSBase):

    def __init__(self, fstype, format_options):
        super(FSExt, self).__init__(fstype, format_options)

    def format(self, device_path, timeout):
        format_options = shlex.split(self.format_options)
        format_options.append(device_path)
        try:
            utils.execute_with_timeout(
                "mkfs", "--type", self.fstype, *format_options,
                timeout=timeout, run_as_root=True, root_helper="sudo")
        except exception.ProcessExecutionError:
            log_fmt = "Could not format '%s'."
            exc_fmt = _("Could not format '%s'.")
            log_and_raise(log_fmt, exc_fmt, device_path)

    def check_format(self, device_path):
        try:
            stdout, stderr = utils.execute(
                "dumpe2fs", device_path, run_as_root=True, root_helper="sudo")
            if 'has_journal' not in stdout:
                msg = _("Volume '%s' does not appear to be formatted.") % (
                    device_path)
                raise exception.GuestError(original_message=msg)
        except exception.ProcessExecutionError as pe:
            if 'Wrong magic number' in pe.stderr:
                volume_fstype = self.fstype
                log_fmt = "'Device '%(dev)s' did not seem to be '%(type)s'."
                exc_fmt = _("'Device '%(dev)s' did not seem to be '%(type)s'.")
                log_and_raise(log_fmt, exc_fmt, {'dev': device_path,
                                                 'type': volume_fstype})
            log_fmt = "Volume '%s' was not formatted."
            exc_fmt = _("Volume '%s' was not formatted.")
            log_and_raise(log_fmt, exc_fmt, device_path)

    def resize(self, device_path, online=False):
        if not online:
            utils.execute("e2fsck", "-f", "-p", device_path,
                          run_as_root=True, root_helper="sudo")
        utils.execute("resize2fs", device_path,
                      run_as_root=True, root_helper="sudo")


class FSExt3(FSExt):

    def __init__(self, format_options):
        super(FSExt3, self).__init__('ext3', format_options)


class FSExt4(FSExt):

    def __init__(self, format_options):
        super(FSExt4, self).__init__('ext4', format_options)


class FSXFS(FSBase):

    def __init__(self, format_options):
        super(FSXFS, self).__init__('xfs', format_options)

    def format(self, device_path, timeout):
        format_options = shlex.split(self.format_options)
        format_options.append(device_path)
        try:
            utils.execute_with_timeout(
                "mkfs.xfs", *format_options,
                timeout=timeout, run_as_root=True, root_helper="sudo")
        except exception.ProcessExecutionError:
            log_fmt = "Could not format '%s'."
            exc_fmt = _("Could not format '%s'.")
            log_and_raise(log_fmt, exc_fmt, device_path)

    def check_format(self, device_path):
        stdout, stderr = utils.execute(
            "xfs_admin", "-l", device_path,
            run_as_root=True, root_helper="sudo")
        if 'not a valid XFS filesystem' in stdout:
            msg = _("Volume '%s' does not appear to be formatted.") % (
                device_path)
            raise exception.GuestError(original_message=msg)

    def resize(self, device_path, online=False):
        utils.execute("xfs_repair", device_path,
                      run_as_root=True, root_helper="sudo")
        utils.execute("mount", device_path,
                      run_as_root=True, root_helper="sudo")
        utils.execute("xfs_growfs", device_path,
                      run_as_root=True, root_helper="sudo")
        utils.execute("umount", device_path,
                      run_as_root=True, root_helper="sudo")


def VolumeFs(fstype, format_options=''):
    supported_fs = {
        'xfs': FSXFS,
        'ext3': FSExt3,
        'ext4': FSExt4
    }
    return supported_fs[fstype](format_options)


class VolumeDevice(object):

    def __init__(self, device_path):
        self.device_path = device_path
        self.volume_fs = VolumeFs(CONF.volume_fstype,
                                  CONF.format_options)

    def migrate_data(self, source_dir, target_subdir=None):
        """Synchronize the data from the source directory to the new
        volume; optionally to a new sub-directory on the new volume.
        """
        self.mount(TMP_MOUNT_POINT, write_to_fstab=False)
        if not source_dir[-1] == '/':
            source_dir = "%s/" % source_dir
        target_dir = TMP_MOUNT_POINT
        if target_subdir:
            target_dir = target_dir + "/" + target_subdir
        try:
            utils.execute("rsync", "--safe-links", "--perms",
                          "--recursive", "--owner", "--group", "--xattrs",
                          "--sparse", source_dir, target_dir,
                          run_as_root=True, root_helper="sudo")
        except exception.ProcessExecutionError:
            log_msg = "Could not migrate data."
            exc_msg = _("Could not migrate date.")
            log_and_raise(log_msg, exc_msg)
        self.unmount(TMP_MOUNT_POINT)

    def _check_device_exists(self):
        """Check that the device path exists.

        Verify that the device path has actually been created and can report
        its size, only then can it be available for formatting, retry
        num_tries to account for the time lag.
        """
        try:
            num_tries = CONF.num_tries
            LOG.debug("Checking if %s exists.", self.device_path)

            utils.execute("blockdev", "--getsize64", self.device_path,
                          run_as_root=True, root_helper="sudo",
                          attempts=num_tries)
        except exception.ProcessExecutionError:
            log_fmt = "Device '%s' is not ready."
            exc_fmt = _("Device '%s' is not ready.")
            log_and_raise(log_fmt, exc_fmt, self.device_path)

    def _check_format(self):
        """Checks that a volume is formatted."""
        LOG.debug("Checking whether '%s' is formatted.", self.device_path)
        self.volume_fs.check_format(self.device_path)

    def _format(self):
        """Calls mkfs to format the device at device_path."""
        LOG.debug("Formatting '%s'.", self.device_path)
        self.volume_fs.format(self.device_path, CONF.volume_format_timeout)

    def format(self):
        """Formats the device at device_path and checks the filesystem."""
        self._check_device_exists()

        try:
            self._check_format()
            LOG.debug(f"Device {self.device_path} already formatted.")
            return
        except exception.GuestError:
            self._format()
            self._check_format()

    def mount(self, mount_point, write_to_fstab=True):
        """Mounts, and writes to fstab."""
        LOG.debug("Will mount %(path)s at %(mount_point)s.",
                  {'path': self.device_path, 'mount_point': mount_point})

        mount_point = VolumeMountPoint(self.device_path, mount_point)
        mount_point.mount()
        if write_to_fstab:
            mount_point.write_to_fstab()

    def _wait_for_mount(self, mount_point, timeout=2):
        """Wait for a fs to be mounted."""
        def wait_for_mount():
            return operating_system.is_mount(mount_point)

        try:
            utils.poll_until(wait_for_mount, sleep_time=1, time_out=timeout)
        except exception.PollTimeOut:
            return False

        return True

    def resize_fs(self, mount_point, online=False):
        """Resize the filesystem on the specified device."""
        self._check_device_exists()
        # Some OS's will mount a file systems after it's attached if
        # an entry is put in the fstab file (like Trove does).
        # Thus it may be necessary to wait for the mount and then unmount
        # the fs again (since the volume was just attached).
        if not online and self._wait_for_mount(mount_point, timeout=2):
            LOG.debug("Unmounting '%s' before resizing.", mount_point)
            self.unmount(mount_point)
        try:
            self.volume_fs.resize(self.device_path, online=online)
        except exception.ProcessExecutionError:
            log_fmt = "Error resizing the filesystem with device '%s'."
            exc_fmt = _("Error resizing the filesystem with device '%s'.")
            log_and_raise(log_fmt, exc_fmt, self.device_path)

    def unmount(self, mount_point):
        if operating_system.is_mount(mount_point):
            try:
                utils.execute("umount", mount_point,
                              run_as_root=True, root_helper='sudo')
            except exception.ProcessExecutionError:
                log_fmt = "Error unmounting '%s'."
                exc_fmt = _("Error unmounting '%s'.")
                log_and_raise(log_fmt, exc_fmt, mount_point)
        else:
            LOG.debug("'%s' is not a mounted fs, cannot unmount", mount_point)

    def unmount_device(self, device_path):
        # unmount if device is already mounted
        mount_points = self.mount_points(device_path)
        for mnt in mount_points:
            LOG.info("Device '%(device)s' is mounted on "
                     "'%(mount_point)s'. Unmounting now.",
                     {'device': device_path, 'mount_point': mnt})
            self.unmount(mnt)

    def mount_points(self, device_path):
        """Returns a list of mount points on the specified device."""
        stdout, stderr = utils.execute(
            "grep '^%s ' /etc/mtab" % device_path,
            shell=True, check_exit_code=[0, 1])
        return [entry.strip().split()[1] for entry in stdout.splitlines()]

    def set_readahead_size(self, readahead_size):
        """Set the readahead size of disk."""
        self._check_device_exists()
        try:
            utils.execute("blockdev", "--setra",
                          readahead_size, self.device_path,
                          run_as_root=True, root_helper="sudo")
        except exception.ProcessExecutionError:
            log_fmt = ("Error setting readahead size to %(size)s "
                       "for device %(device)s.")
            exc_fmt = _("Error setting readahead size to %(size)s "
                        "for device %(device)s.")
            log_and_raise(log_fmt, exc_fmt, {'size': readahead_size,
                                             'device': self.device_path})


class VolumeMountPoint(object):

    def __init__(self, device_path, mount_point):
        self.device_path = device_path
        self.mount_point = mount_point
        self.volume_fstype = CONF.volume_fstype
        self.mount_options = CONF.mount_options

    def mount(self):
        if not operating_system.exists(self.mount_point, is_directory=True,
                                       as_root=True):
            operating_system.ensure_directory(self.mount_point, as_root=True)
        LOG.debug("Mounting volume. Device path:{0}, mount_point:{1}, "
                  "volume_type:{2}, mount options:{3}".format(
                      self.device_path, self.mount_point, self.volume_fstype,
                      self.mount_options))
        try:
            utils.execute("mount", "-t", self.volume_fstype,
                          "-o", self.mount_options,
                          self.device_path, self.mount_point,
                          run_as_root=True, root_helper="sudo")
        except exception.ProcessExecutionError:
            log_fmt = "Could not mount '%s'."
            exc_fmt = _("Could not mount '%s'.")
            log_and_raise(log_fmt, exc_fmt, self.mount_point)

    def write_to_fstab(self):
        fstab_line = ("%s\t%s\t%s\t%s\t0\t0" %
                      (self.device_path, self.mount_point, self.volume_fstype,
                       self.mount_options))
        LOG.debug("Writing new line to fstab:%s", fstab_line)
        with open('/etc/fstab', "r") as fstab:
            fstab_content = fstab.read()
        with NamedTemporaryFile(mode='w', delete=False) as tempfstab:
            tempfstab.write(fstab_content + fstab_line)
        try:
            utils.execute("install", "-o", "root", "-g", "root",
                          "-m", "644", tempfstab.name, "/etc/fstab",
                          run_as_root=True, root_helper="sudo")
        except exception.ProcessExecutionError:
            log_fmt = "Could not add '%s' to fstab."
            exc_fmt = _("Could not add '%s' to fstab.")
            log_and_raise(log_fmt, exc_fmt, self.mount_point)
        os.remove(tempfstab.name)



# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import proboscis

from trove.tests.scenario.runners.test_runners import TestRunner


class InstanceDeleteRunner(TestRunner):

    def __init__(self):
        super(InstanceDeleteRunner, self).__init__()

    def run_instance_delete(self, expected_http_code=202):
        if self.has_do_not_delete_instance:
            self.report.log("TESTS_DO_NOT_DELETE_INSTANCE=True was "
                            "specified, skipping delete...")
            raise proboscis.SkipTest("TESTS_DO_NOT_DELETE_INSTANCE "
                                     "was specified.")

        self.assert_instance_delete(self.instance_info.id, expected_http_code)

    def assert_instance_delete(self, instance_id, expected_http_code):
        self.report.log("Testing delete on instance: %s" % instance_id)

        client = self.auth_client
        client.instances.delete(instance_id)
        self.assert_client_code(client, expected_http_code)

    def run_instance_delete_wait(self, expected_states=['SHUTDOWN']):
        if self.has_do_not_delete_instance:
            self.report.log("TESTS_DO_NOT_DELETE_INSTANCE=True was "
                            "specified, skipping delete wait...")
            raise proboscis.SkipTest("TESTS_DO_NOT_DELETE_INSTANCE "
                                     "was specified.")
        self.assert_all_gone(self.instance_info.id, expected_states[-1])
        self.assert_server_group_gone(self.instance_info.srv_grp_id)



# Copyright 2010-2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from trove.common import wsgi


def create_links(resource_path, request, id):
    """Creates the links dictionary in the format typical of most resources."""
    context = request.environ[wsgi.CONTEXT_KEY]
    link_info = {
        'host': request.host,
        'version': request.url_version,
        'tenant_id': context.project_id,
        'resource_path': resource_path,
        'id': id,
    }
    return [
        {
            "href": "https://%(host)s/v%(version)s/%(tenant_id)s"
                    "/%(resource_path)s/%(id)s" % link_info,
            "rel": "self"
        },
        {
            "href": "https://%(host)s/%(resource_path)s/%(id)s" % link_info,
            "rel": "bookmark"
        }
    ]



# Copyright [2015] Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging

from trove.backup import models as backup_model
from trove.common import apischema
from trove.common import clients
from trove.common import exception
from trove.common import glance as common_glance
from trove.common import utils
from trove.common import wsgi
from trove.common.auth import admin_context
from trove.configuration import models as config_model
from trove.datastore import models
from trove.extensions.mgmt.datastores import views
from trove.instance import models as instance_model

LOG = logging.getLogger(__name__)


class DatastoreVersionController(wsgi.Controller):
    """Controller for datastore version registration functionality."""

    schemas = apischema.mgmt_datastore_version

    @admin_context
    def create(self, req, body, tenant_id):
        """Adds a new datastore version."""
        context = req.environ[wsgi.CONTEXT_KEY]
        datastore_name = body['version']['datastore_name']
        version_name = body['version']['name']
        manager = body['version']['datastore_manager']
        image_id = body['version'].get('image')
        image_tags = body['version'].get('image_tags')
        packages = body['version'].get('packages')
        if type(packages) is list:
            packages = ','.join(packages)
        active = body['version']['active']
        default = body['version'].get('default', False)
        # For backward compatibility, use name as default value for version if
        # not specified
        version_str = body['version'].get('version', version_name)

        LOG.info("Tenant: '%(tenant)s' is adding the datastore "
                 "version: '%(version)s' to datastore: '%(datastore)s'",
                 {'tenant': tenant_id, 'version': version_name,
                  'datastore': datastore_name})

        if not image_id and not image_tags:
            raise exception.BadRequest("Image must be specified.")

        client = clients.create_glance_client(context)
        common_glance.get_image_id(client, image_id, image_tags)

        try:
            datastore = models.Datastore.load(datastore_name)
        except exception.DatastoreNotFound:
            # Create the datastore if datastore_name does not exists.
            LOG.info("Creating datastore %s", datastore_name)
            datastore = models.DBDatastore()
            datastore.id = utils.generate_uuid()
            datastore.name = datastore_name
            datastore.save()

        try:
            models.DatastoreVersion.load(datastore, version_name,
                                         version=version_str)
            raise exception.DatastoreVersionAlreadyExists(
                name=version_name, version=version_str)
        except exception.DatastoreVersionNotFound:
            models.update_datastore_version(datastore.name, version_name,
                                            manager, image_id, image_tags,
                                            packages, active,
                                            version=version_str)

        if default:
            models.update_datastore(datastore.name, version_name)

        return wsgi.Result(None, 202)

    @admin_context
    def index(self, req, tenant_id):
        """Lists all datastore-versions for given datastore."""
        db_ds_versions = models.DatastoreVersions.load_all(only_active=False)
        datastore_versions = [models.DatastoreVersion.load_by_uuid(
            ds_version.id) for ds_version in db_ds_versions]

        return wsgi.Result(
            views.DatastoreVersionsView(datastore_versions).data(), 200)

    @admin_context
    def show(self, req, tenant_id, id):
        """Lists details of a datastore-version for given datastore."""
        datastore_version = models.DatastoreVersion.load_by_uuid(id)
        return wsgi.Result(
            views.DatastoreVersionView(datastore_version).data(),
            200)

    @admin_context
    def edit(self, req, body, tenant_id, id):
        """Updates the attributes of a datastore version."""
        context = req.environ[wsgi.CONTEXT_KEY]
        datastore_version = models.DatastoreVersion.load_by_uuid(id)

        LOG.info("Tenant: '%(tenant)s' is updating the datastore "
                 "version: '%(id)s' for datastore: '%(datastore)s'",
                 {'tenant': tenant_id, 'id': id,
                  'datastore': datastore_version.datastore_name})

        name = body.get('name', datastore_version.name)
        manager = body.get('datastore_manager', datastore_version.manager)
        image_id = body.get('image')
        image_tags = body.get('image_tags')
        active = body.get('active', datastore_version.active)
        default = body.get('default', None)
        packages = body.get('packages', datastore_version.packages)
        if type(packages) is list:
            packages = ','.join(packages)

        if image_id or image_tags:
            client = clients.create_glance_client(context)
            common_glance.get_image_id(client, image_id, image_tags)

        if not image_id and image_tags:
            # Remove the image ID from the datastore version.
            image_id = ""

        if image_id is None:
            image_id = datastore_version.image_id
        if image_tags is None:
            image_tags = datastore_version.image_tags
            if type(image_tags) is str:
                image_tags = image_tags.split(',')

        if not image_id and not image_tags:
            raise exception.BadRequest("Image must be specified.")

        models.update_datastore_version(datastore_version.datastore_name,
                                        datastore_version.name,
                                        manager, image_id, image_tags,
                                        packages, active,
                                        version=datastore_version.version,
                                        new_name=name)

        if default:
            models.update_datastore(datastore_version.datastore_name,
                                    datastore_version.name)
        elif (default is False and datastore_version.default is True):
            models.update_datastore(datastore_version.datastore_name, None)

        return wsgi.Result(None, 202)

    @admin_context
    def delete(self, req, tenant_id, id):
        """Remove an existing datastore version."""
        instances = instance_model.DBInstance.find_all(
            datastore_version_id=id, deleted=0).all()
        if len(instances) > 0:
            raise exception.DatastoreVersionsInUse(resource='instance')

        backups = backup_model.DBBackup.find_all(
            datastore_version_id=id, deleted=0).all()
        if len(backups) > 0:
            raise exception.DatastoreVersionsInUse(resource='backup')

        configs = config_model.DBConfiguration.find_all(
            datastore_version_id=id, deleted=0).all()
        if len(configs) > 0:
            raise exception.DatastoreVersionsInUse(resource='configuration')

        datastore_version = models.DatastoreVersion.load_by_uuid(id)
        datastore = models.Datastore.load(datastore_version.datastore_id)

        LOG.info("Tenant: '%(tenant)s' is removing the datastore "
                 "version: '%(version)s' for datastore: '%(datastore)s'",
                 {'tenant': tenant_id, 'version': datastore_version.name,
                  'datastore': datastore.name})

        # Remove the config parameters associated with the datastore version
        LOG.debug(f"Deleting config parameters for datastore version {id}")
        db_params = config_model.DatastoreConfigurationParameters. \
            load_parameters(id)
        for db_param in db_params:
            db_param.delete()

        if datastore.default_version_id == datastore_version.id:
            models.update_datastore(datastore.name, None)
        datastore_version.delete()
        return wsgi.Result(None, 202)



# Copyright 2016 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


charset = {"big5": ["big5_chinese_ci", "big5_bin"],
           "dec8": ["dec8_swedish_ci", "dec8_bin"],
           "cp850": ["cp850_general_ci", "cp850_bin"],
           "hp8": ["hp8_english_ci", "hp8_bin"],
           "koi8r": ["koi8r_general_ci", "koi8r_bin"],
           "latin1": ["latin1_swedish_ci",
                      "latin1_german1_ci",
                      "latin1_danish_ci",
                      "latin1_german2_ci",
                      "latin1_bin",
                      "latin1_general_ci",
                      "latin1_general_cs",
                      "latin1_spanish_ci"],
           "latin2": ["latin2_general_ci",
                      "latin2_czech_cs",
                      "latin2_hungarian_ci",
                      "latin2_croatian_ci",
                      "latin2_bin"],
           "swe7": ["swe7_swedish_ci", "swe7_bin"],
           "ascii": ["ascii_general_ci", "ascii_bin"],
           "ujis": ["ujis_japanese_ci", "ujis_bin"],
           "sjis": ["sjis_japanese_ci", "sjis_bin"],
           "hebrew": ["hebrew_general_ci", "hebrew_bin"],
           "tis620": ["tis620_thai_ci", "tis620_bin"],
           "euckr": ["euckr_korean_ci", "euckr_bin"],
           "koi8u": ["koi8u_general_ci", "koi8u_bin"],
           "gb2312": ["gb2312_chinese_ci", "gb2312_bin"],
           "greek": ["greek_general_ci", "greek_bin"],
           "cp1250": ["cp1250_general_ci",
                      "cp1250_czech_cs",
                      "cp1250_croatian_ci",
                      "cp1250_bin",
                      "cp1250_polish_ci"],
           "gbk": ["gbk_chinese_ci", "gbk_bin"],
           "latin5": ["latin5_turkish_ci", "latin5_bin"],
           "armscii8": ["armscii8_general_ci", "armscii8_bin"],
           "utf8": ["utf8_general_ci",
                    "utf8_bin",
                    "utf8_unicode_ci",
                    "utf8_icelandic_ci",
                    "utf8_latvian_ci",
                    "utf8_romanian_ci",
                    "utf8_slovenian_ci",
                    "utf8_polish_ci",
                    "utf8_estonian_ci",
                    "utf8_spanish_ci",
                    "utf8_swedish_ci",
                    "utf8_turkish_ci",
                    "utf8_czech_ci",
                    "utf8_danish_ci",
                    "utf8_lithuanian_ci",
                    "utf8_slovak_ci",
                    "utf8_spanish2_ci",
                    "utf8_roman_ci",
                    "utf8_persian_ci",
                    "utf8_esperanto_ci",
                    "utf8_hungarian_ci",
                    "utf8_sinhala_ci",
                    "utf8_german2_ci",
                    "utf8_croatian_ci",
                    "utf8_unicode_520_ci",
                    "utf8_vietnamese_ci",
                    "utf8_general_mysql500_ci"
                    ],
           "utf8mb4": ["utf8mb4_0900_ai_ci"],
           "ucs2": ["ucs2_general_ci",
                    "ucs2_bin",
                    "ucs2_unicode_ci",
                    "ucs2_icelandic_ci",
                    "ucs2_latvian_ci",
                    "ucs2_romanian_ci",
                    "ucs2_slovenian_ci",
                    "ucs2_polish_ci",
                    "ucs2_estonian_ci",
                    "ucs2_spanish_ci",
                    "ucs2_swedish_ci",
                    "ucs2_turkish_ci",
                    "ucs2_czech_ci",
                    "ucs2_danish_ci",
                    "ucs2_lithuanian_ci",
                    "ucs2_slovak_ci",
                    "ucs2_spanish2_ci",
                    "ucs2_roman_ci",
                    "ucs2_persian_ci",
                    "ucs2_esperanto_ci",
                    "ucs2_hungarian_ci",
                    "ucs2_sinhala_ci",
                    "ucs2_german2_ci",
                    "ucs2_croatian_ci",
                    "ucs2_unicode_520_ci",
                    "ucs2_vietnamese_ci",
                    "ucs2_general_mysql500_ci"
                    ],
           "cp866": ["cp866_general_ci", "cp866_bin"],
           "keybcs2": ["keybcs2_general_ci", "keybcs2_bin"],
           "macce": ["macce_general_ci", "macce_bin"],
           "macroman": ["macroman_general_ci", "macroman_bin"],
           "cp852": ["cp852_general_ci", "cp852_bin"],
           "latin7": ["latin7_general_ci",
                      "latin7_estonian_cs",
                      "latin7_general_cs",
                      "latin7_bin"],
           "utf8mb4": ["utf8mb4_general_ci",
                       "utf8mb4_bin",
                       "utf8mb4_unicode_ci",
                       "utf8mb4_icelandic_ci",
                       "utf8mb4_latvian_ci",
                       "utf8mb4_romanian_ci",
                       "utf8mb4_slovenian_ci",
                       "utf8mb4_polish_ci",
                       "utf8mb4_estonian_ci",
                       "utf8mb4_spanish_ci",
                       "utf8mb4_swedish_ci",
                       "utf8mb4_turkish_ci",
                       "utf8mb4_czech_ci",
                       "utf8mb4_danish_ci",
                       "utf8mb4_lithuanian_ci",
                       "utf8mb4_slovak_ci",
                       "utf8mb4_spanish2_ci",
                       "utf8mb4_roman_ci",
                       "utf8mb4_persian_ci",
                       "utf8mb4_esperanto_ci",
                       "utf8mb4_hungarian_ci",
                       "utf8mb4_sinhala_ci",
                       "utf8mb4_german2_ci",
                       "utf8mb4_croatian_ci",
                       "utf8mb4_unicode_520_ci",
                       "utf8mb4_vietnamese_ci"],
           "cp1251": ["cp1251_general_ci",
                      "cp1251_bulgarian_ci",
                      "cp1251_ukrainian_ci",
                      "cp1251_bin",
                      "cp1251_general_cs"],
           "utf16": ["utf16_general_ci",
                     "utf16_bin",
                     "utf16_unicode_ci",
                     "utf16_icelandic_ci",
                     "utf16_latvian_ci",
                     "utf16_romanian_ci",
                     "utf16_slovenian_ci",
                     "utf16_polish_ci",
                     "utf16_estonian_ci",
                     "utf16_spanish_ci",
                     "utf16_swedish_ci",
                     "utf16_turkish_ci",
                     "utf16_czech_ci",
                     "utf16_danish_ci",
                     "utf16_lithuanian_ci",
                     "utf16_slovak_ci",
                     "utf16_spanish2_ci",
                     "utf16_roman_ci",
                     "utf16_persian_ci",
                     "utf16_esperanto_ci",
                     "utf16_hungarian_ci",
                     "utf16_sinhala_ci",
                     "utf16_german2_ci",
                     "utf16_croatian_ci",
                     "utf16_unicode_520_ci",
                     "utf16_vietnamese_ci"],
           "utf16le": ["utf16le_general_ci",
                       "utf16le_bin"],
           "cp1256": ["cp1256_general_ci", "cp1256_bin"],
           "cp1257": ["cp1257_general_ci",
                      "cp1257_lithuanian_ci",
                      "cp1257_bin"],
           "utf32": ["utf32_general_ci",
                     "utf32_bin",
                     "utf32_unicode_ci",
                     "utf32_icelandic_ci",
                     "utf32_latvian_ci",
                     "utf32_romanian_ci",
                     "utf32_slovenian_ci",
                     "utf32_polish_ci",
                     "utf32_estonian_ci",
                     "utf32_spanish_ci",
                     "utf32_swedish_ci",
                     "utf32_turkish_ci",
                     "utf32_czech_ci",
                     "utf32_danish_ci",
                     "utf32_lithuanian_ci",
                     "utf32_slovak_ci",
                     "utf32_spanish2_ci",
                     "utf32_roman_ci",
                     "utf32_persian_ci",
                     "utf32_esperanto_ci",
                     "utf32_hungarian_ci",
                     "utf32_sinhala_ci",
                     "utf32_german2_ci",
                     "utf32_croatian_ci",
                     "utf32_unicode_520_ci",
                     "utf32_vietnamese_ci"],
           "binary": ["binary"],
           "geostd8": ["geostd8_general_ci", "geostd8_bin"],
           "cp932": ["cp932_japanese_ci", "cp932_bin"],
           "eucjpms": ["eucjpms_japanese_ci", "eucjpms_bin"],
           "gb18030": ["gb18030_chinese_ci",
                       "gb18030_bin",
                       "gb18030_unicode_520_ci"]}

collation = {"big5_chinese_ci": "big5",
             "big5_bin": "big5",
             "dec8_swedish_ci": "dec8",
             "dec8_bin": "dec8",
             "cp850_general_ci": "cp850",
             "cp850_bin": "cp850",
             "hp8_english_ci": "hp8",
             "hp8_bin": "hp8",
             "koi8r_general_ci": "koi8r",
             "koi8r_bin": "koi8r",
             "latin1_german1_ci": "latin1",
             "latin1_swedish_ci": "latin1",
             "latin1_danish_ci": "latin1",
             "latin1_german2_ci": "latin1",
             "latin1_bin": "latin1",
             "latin1_general_ci": "latin1",
             "latin1_general_cs": "latin1",
             "latin1_spanish_ci": "latin1",
             "latin2_czech_cs": "latin2",
             "latin2_general_ci": "latin2",
             "latin2_hungarian_ci": "latin2",
             "latin2_croatian_ci": "latin2",
             "latin2_bin": "latin2",
             "swe7_swedish_ci": "swe7",
             "swe7_bin": "swe7",
             "ascii_general_ci": "ascii",
             "ascii_bin": "ascii",
             "ujis_japanese_ci": "ujis",
             "ujis_bin": "ujis",
             "sjis_japanese_ci": "sjis",
             "sjis_bin": "sjis",
             "hebrew_general_ci": "hebrew",
             "hebrew_bin": "hebrew",
             "tis620_thai_ci": "tis620",
             "tis620_bin": "tis620",
             "euckr_korean_ci": "euckr",
             "euckr_bin": "euckr",
             "koi8u_general_ci": "koi8u",
             "koi8u_bin": "koi8u",
             "gb2312_chinese_ci": "gb2312",
             "gb2312_bin": "gb2312",
             "greek_general_ci": "greek",
             "greek_bin": "greek",
             "cp1250_general_ci": "cp1250",
             "cp1250_czech_cs": "cp1250",
             "cp1250_croatian_ci": "cp1250",
             "cp1250_bin": "cp1250",
             "cp1250_polish_ci": "cp1250",
             "gbk_chinese_ci": "gbk",
             "gbk_bin": "gbk",
             "latin5_turkish_ci": "latin5",
             "latin5_bin": "latin5",
             "armscii8_general_ci": "armscii8",
             "armscii8_bin": "armscii8",
             "utf8_general_ci": "utf8",
             "utf8_bin": "utf8",
             "utf8_unicode_ci": "utf8",
             "utf8_icelandic_ci": "utf8",
             "utf8_latvian_ci": "utf8",
             "utf8_romanian_ci": "utf8",
             "utf8_slovenian_ci": "utf8",
             "utf8_polish_ci": "utf8",
             "utf8_estonian_ci": "utf8",
             "utf8_spanish_ci": "utf8",
             "utf8_swedish_ci": "utf8",
             "utf8_turkish_ci": "utf8",
             "utf8_czech_ci": "utf8",
             "utf8_danish_ci": "utf8",
             "utf8_lithuanian_ci": "utf8",
             "utf8_slovak_ci": "utf8",
             "utf8_spanish2_ci": "utf8",
             "utf8_roman_ci": "utf8",
             "utf8_persian_ci": "utf8",
             "utf8_esperanto_ci": "utf8",
             "utf8_hungarian_ci": "utf8",
             "utf8_sinhala_ci": "utf8",
             "utf8_german2_ci": "utf8",
             "utf8_croatian_ci": "utf8",
             "utf8_unicode_520_ci": "utf8",
             "utf8_vietnamese_ci": "utf8",
             "utf8_general_mysql500_ci": "utf8",
             "utf8mb4_0900_ai_ci": "utf8mb4",
             "ucs2_general_ci": "ucs2",
             "ucs2_bin": "ucs2",
             "ucs2_unicode_ci": "ucs2",
             "ucs2_icelandic_ci": "ucs2",
             "ucs2_latvian_ci": "ucs2",
             "ucs2_romanian_ci": "ucs2",
             "ucs2_slovenian_ci": "ucs2",
             "ucs2_polish_ci": "ucs2",
             "ucs2_estonian_ci": "ucs2",
             "ucs2_spanish_ci": "ucs2",
             "ucs2_swedish_ci": "ucs2",
             "ucs2_turkish_ci": "ucs2",
             "ucs2_czech_ci": "ucs2",
             "ucs2_danish_ci": "ucs2",
             "ucs2_lithuanian_ci": "ucs2",
             "ucs2_slovak_ci": "ucs2",
             "ucs2_spanish2_ci": "ucs2",
             "ucs2_roman_ci": "ucs2",
             "ucs2_persian_ci": "ucs2",
             "ucs2_esperanto_ci": "ucs2",
             "ucs2_hungarian_ci": "ucs2",
             "ucs2_sinhala_ci": "ucs2",
             "ucs2_german2_ci": "ucs2",
             "ucs2_croatian_ci": "ucs2",
             "ucs2_unicode_520_ci": "ucs2",
             "ucs2_vietnamese_ci": "ucs2",
             "ucs2_general_mysql500_ci": "ucs2",
             "cp866_general_ci": "cp866",
             "cp866_bin": "cp866",
             "keybcs2_general_ci": "keybcs2",
             "keybcs2_bin": "keybcs2",
             "macce_general_ci": "macce",
             "macce_bin": "macce",
             "macroman_general_ci": "macroman",
             "macroman_bin": "macroman",
             "cp852_general_ci": "cp852",
             "cp852_bin": "cp852",
             "latin7_estonian_cs": "latin7",
             "latin7_general_ci": "latin7",
             "latin7_general_cs": "latin7",
             "latin7_bin": "latin7",
             "utf8mb4_general_ci": "utf8mb4",
             "utf8mb4_bin": "utf8mb4",
             "utf8mb4_unicode_ci": "utf8mb4",
             "utf8mb4_icelandic_ci": "utf8mb4",
             "utf8mb4_latvian_ci": "utf8mb4",
             "utf8mb4_romanian_ci": "utf8mb4",
             "utf8mb4_slovenian_ci": "utf8mb4",
             "utf8mb4_polish_ci": "utf8mb4",
             "utf8mb4_estonian_ci": "utf8mb4",
             "utf8mb4_spanish_ci": "utf8mb4",
             "utf8mb4_swedish_ci": "utf8mb4",
             "utf8mb4_turkish_ci": "utf8mb4",
             "utf8mb4_czech_ci": "utf8mb4",
             "utf8mb4_danish_ci": "utf8mb4",
             "utf8mb4_lithuanian_ci": "utf8mb4",
             "utf8mb4_slovak_ci": "utf8mb4",
             "utf8mb4_spanish2_ci": "utf8mb4",
             "utf8mb4_roman_ci": "utf8mb4",
             "utf8mb4_persian_ci": "utf8mb4",
             "utf8mb4_esperanto_ci": "utf8mb4",
             "utf8mb4_hungarian_ci": "utf8mb4",
             "utf8mb4_sinhala_ci": "utf8mb4",
             "utf8mb4_german2_ci": "utf8mb4",
             "utf8mb4_croatian_ci": "utf8mb4",
             "utf8mb4_unicode_520_ci": "utf8mb4",
             "utf8mb4_vietnamese_ci": "utf8mb4",
             "cp1251_bulgarian_ci": "cp1251",
             "cp1251_ukrainian_ci": "cp1251",
             "cp1251_bin": "cp1251",
             "cp1251_general_ci": "cp1251",
             "cp1251_general_cs": "cp1251",
             "utf16_general_ci": "utf16",
             "utf16_bin": "utf16",
             "utf16_unicode_ci": "utf16",
             "utf16_icelandic_ci": "utf16",
             "utf16_latvian_ci": "utf16",
             "utf16_romanian_ci": "utf16",
             "utf16_slovenian_ci": "utf16",
             "utf16_polish_ci": "utf16",
             "utf16_estonian_ci": "utf16",
             "utf16_spanish_ci": "utf16",
             "utf16_swedish_ci": "utf16",
             "utf16_turkish_ci": "utf16",
             "utf16_czech_ci": "utf16",
             "utf16_danish_ci": "utf16",
             "utf16_lithuanian_ci": "utf16",
             "utf16_slovak_ci": "utf16",
             "utf16_spanish2_ci": "utf16",
             "utf16_roman_ci": "utf16",
             "utf16_persian_ci": "utf16",
             "utf16_esperanto_ci": "utf16",
             "utf16_hungarian_ci": "utf16",
             "utf16_sinhala_ci": "utf16",
             "utf16_german2_ci": "utf16",
             "utf16_croatian_ci": "utf16",
             "utf16_unicode_520_ci": "utf16",
             "utf16_vietnamese_ci": "utf16",
             "utf16le_general_ci": "utf16le",
             "utf16le_bin": "utf16le",
             "cp1256_general_ci": "cp1256",
             "cp1256_bin": "cp1256",
             "cp1257_lithuanian_ci": "cp1257",
             "cp1257_bin": "cp1257",
             "cp1257_general_ci": "cp1257",
             "utf32_general_ci": "utf32",
             "utf32_bin": "utf32",
             "utf32_unicode_ci": "utf32",
             "utf32_icelandic_ci": "utf32",
             "utf32_latvian_ci": "utf32",
             "utf32_romanian_ci": "utf32",
             "utf32_slovenian_ci": "utf32",
             "utf32_polish_ci": "utf32",
             "utf32_estonian_ci": "utf32",
             "utf32_spanish_ci": "utf32",
             "utf32_swedish_ci": "utf32",
             "utf32_turkish_ci": "utf32",
             "utf32_czech_ci": "utf32",
             "utf32_danish_ci": "utf32",
             "utf32_lithuanian_ci": "utf32",
             "utf32_slovak_ci": "utf32",
             "utf32_spanish2_ci": "utf32",
             "utf32_roman_ci": "utf32",
             "utf32_persian_ci": "utf32",
             "utf32_esperanto_ci": "utf32",
             "utf32_hungarian_ci": "utf32",
             "utf32_sinhala_ci": "utf32",
             "utf32_german2_ci": "utf32",
             "utf32_croatian_ci": "utf32",
             "utf32_unicode_520_ci": "utf32",
             "utf32_vietnamese_ci": "utf32",
             "binary": "binary",
             "geostd8_general_ci": "geostd8",
             "geostd8_bin": "geostd8",
             "cp932_japanese_ci": "cp932",
             "cp932_bin": "cp932",
             "eucjpms_japanese_ci": "eucjpms",
             "eucjpms_bin": "eucjpms",
             "gb18030_chinese_ci": "gb18030",
             "gb18030_bin": "gb18030",
             "gb18030_unicode_520_ci": "gb18030"}



# Copyright 2020 Catalyst Cloud
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import re

from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log as logging

from backup.drivers import mysql_base

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class InnoBackupEx(mysql_base.MySQLBaseRunner):
    """Implementation of Backup and Restore for InnoBackupEx."""
    backup_log = '/tmp/innobackupex.log'
    prepare_log = '/tmp/prepare.log'
    restore_cmd = ('xbstream -x -C %(restore_location)s --parallel=2'
                   ' 2>/tmp/xbstream_extract.log')
    prepare_cmd = ('innobackupex'
                   ' --defaults-file=%(restore_location)s/backup-my.cnf'
                   ' --ibbackup=xtrabackup'
                   ' --apply-log'
                   ' %(restore_location)s'
                   ' 2>' + prepare_log)

    @property
    def cmd(self):
        cmd = ('innobackupex'
               ' --stream=xbstream'
               ' --parallel=2 ' +
               self.user_and_pass + ' %s' % self.datadir +
               ' 2>' + self.backup_log
               )
        return cmd + self.zip_cmd + self.encrypt_cmd

    def check_restore_process(self):
        """Check whether xbstream restore is successful."""
        LOG.info('Checking return code of xbstream restore process.')
        return_code = self.process.wait()
        if return_code != 0:
            LOG.error('xbstream exited with %s', return_code)
            return False

        with open('/tmp/xbstream_extract.log', 'r') as xbstream_log:
            for line in xbstream_log:
                # Ignore empty lines
                if not line.strip():
                    continue

                LOG.error('xbstream restore failed with: %s',
                          line.rstrip('\n'))
                return False

        return True

    def post_restore(self):
        """Hook that is called after the restore command."""
        LOG.info("Running innobackupex prepare: %s.", self.prepare_command)
        processutils.execute(self.prepare_command, shell=True)

        LOG.info("Checking innobackupex prepare log")
        with open(self.prepare_log, 'r') as prepare_log:
            output = prepare_log.read()
            if not output:
                msg = "innobackupex prepare log file empty"
                raise Exception(msg)

            last_line = output.splitlines()[-1].strip()
            if not re.search('completed OK!', last_line):
                msg = "innobackupex prepare did not complete successfully"
                raise Exception(msg)


class InnoBackupExIncremental(InnoBackupEx):
    """InnoBackupEx incremental backup."""

    incremental_prep = ('innobackupex'
                        ' --defaults-file=%(restore_location)s/backup-my.cnf'
                        ' --ibbackup=xtrabackup'
                        ' --apply-log'
                        ' --redo-only'
                        ' %(restore_location)s'
                        ' %(incremental_args)s'
                        ' 2>/tmp/innoprepare.log')

    def __init__(self, *args, **kwargs):
        if not kwargs.get('lsn'):
            raise AttributeError('lsn attribute missing')
        self.parent_location = kwargs.pop('parent_location', '')
        self.parent_checksum = kwargs.pop('parent_checksum', '')

        super(InnoBackupExIncremental, self).__init__(*args, **kwargs)

    @property
    def cmd(self):
        cmd = ('innobackupex'
               ' --stream=xbstream'
               ' --incremental'
               ' --incremental-lsn=%(lsn)s ' +
               self.user_and_pass + ' %s' % self.datadir +
               ' 2>' + self.backup_log)
        return cmd + self.zip_cmd + self.encrypt_cmd

    def get_metadata(self):
        _meta = super(InnoBackupExIncremental, self).get_metadata()

        _meta.update({
            'parent_location': self.parent_location,
            'parent_checksum': self.parent_checksum,
        })
        return _meta

    def run_restore(self):
        """Run incremental restore.

        First grab all parents and prepare them with '--redo-only'. After
        all backups are restored the super class InnoBackupEx post_restore
        method is called to do the final prepare with '--apply-log'
        """
        LOG.debug('Running incremental restore')
        self.incremental_restore(self.location, self.checksum)
        return self.restore_content_length



# Copyright 2016 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest.mock import ANY
from unittest.mock import call
from unittest.mock import DEFAULT
from unittest.mock import MagicMock
from unittest.mock import Mock
from unittest.mock import patch
from unittest.mock import PropertyMock

from neutronclient.common import exceptions as neutron_exceptions

from trove.cluster import models
from trove.common import clients
from trove.common import exception
from trove.tests.unittests import trove_testtools


class TestModels(trove_testtools.TestCase):

    @patch.object(clients, 'create_nova_client', return_value=MagicMock())
    def test_validate_instance_flavors(self, create_nova_cli_mock):
        patch.object(
            create_nova_cli_mock.return_value, 'flavors',
            new_callable=PropertyMock(return_value=Mock()))
        mock_flv = create_nova_cli_mock.return_value.flavors.get.return_value
        mock_flv.ephemeral = 0

        test_instances = [{'flavor_id': 1, 'volume_size': 10},
                          {'flavor_id': 1, 'volume_size': 1.5,
                           'region_name': 'home'},
                          {'flavor_id': 2, 'volume_size': 3,
                           'region_name': 'work'}]
        models.validate_instance_flavors(Mock(), test_instances,
                                         True, True)
        create_nova_cli_mock.assert_has_calls([call(ANY, None),
                                               call(ANY, 'home'),
                                               call(ANY, 'work')])

        self.assertRaises(exception.LocalStorageNotSpecified,
                          models.validate_instance_flavors,
                          Mock(), test_instances, False, True)

        mock_flv.ephemeral = 1
        models.validate_instance_flavors(Mock(), test_instances,
                                         False, True)

    def test_validate_volume_size(self):
        self.patch_conf_property('max_accepted_volume_size', 10)
        models.validate_volume_size(9)
        models.validate_volume_size(10)

        self.assertRaises(exception.VolumeQuotaExceeded,
                          models.validate_volume_size, 11)

        self.assertRaises(exception.VolumeSizeNotSpecified,
                          models.validate_volume_size, None)

    @patch.object(models, 'validate_volume_size')
    def test_get_required_volume_size(self, vol_size_validator_mock):
        test_instances = [{'flavor_id': 1, 'volume_size': 10},
                          {'flavor_id': 1, 'volume_size': 1.5},
                          {'flavor_id': 1, 'volume_size': 3}]
        total_size = models.get_required_volume_size(test_instances, True)
        self.assertEqual(14.5, total_size)
        vol_size_validator_mock.assert_has_calls([call(10),
                                                  call(1.5),
                                                  call(3)], any_order=True)

        test_instances = [{'flavor_id': 1, 'volume_size': 10},
                          {'flavor_id': 1, 'volume_size': 1.5},
                          {'flavor_id': 1, 'volume_size': None}]
        self.assertRaises(exception.ClusterVolumeSizeRequired,
                          models.get_required_volume_size,
                          test_instances, True)

        test_instances = [{'flavor_id': 1, 'volume_size': 10},
                          {'flavor_id': 1, 'volume_size': 1.5},
                          {'flavor_id': 1}]
        self.assertRaises(exception.ClusterVolumeSizeRequired,
                          models.get_required_volume_size,
                          test_instances, True)

        test_instances = [{'flavor_id': 1},
                          {'flavor_id': 1},
                          {'flavor_id': 1}]
        total_size = models.get_required_volume_size(test_instances, False)
        self.assertIsNone(total_size)

        test_instances = [{'flavor_id': 1, 'volume_size': 10},
                          {'flavor_id': 1, 'volume_size': 1.5}]
        self.assertRaises(exception.VolumeNotSupported,
                          models.get_required_volume_size,
                          test_instances, False)

    def test_assert_same_instance_volumes(self):
        test_instances = [{'flavor_id': 1, 'volume_size': 10},
                          {'flavor_id': 1, 'volume_size': 10},
                          {'flavor_id': 1, 'volume_size': 10}]
        models.assert_same_instance_volumes(test_instances)

        test_instances = [{'flavor_id': 1, 'volume_size': 5},
                          {'flavor_id': 1, 'volume_size': 5},
                          {'flavor_id': 1, 'volume_size': 5}]
        models.assert_same_instance_volumes(test_instances, required_size=5)

        test_instances = [{'flavor_id': 1, 'volume_size': 10},
                          {'flavor_id': 1, 'volume_size': 1.5},
                          {'flavor_id': 1, 'volume_size': 10}]
        self.assertRaises(exception.ClusterVolumeSizesNotEqual,
                          models.assert_same_instance_volumes,
                          test_instances)

        test_instances = [{'flavor_id': 1, 'volume_size': 10},
                          {'flavor_id': 1, 'volume_size': 10},
                          {'flavor_id': 1, 'volume_size': 10}]
        self.assertRaises(exception.ClusterVolumeSizesNotEqual,
                          models.assert_same_instance_volumes,
                          test_instances, required_size=5)

    def test_assert_same_instance_flavors(self):
        test_instances = [{'flavor_id': 1, 'volume_size': 10},
                          {'flavor_id': 1, 'volume_size': 10},
                          {'flavor_id': 1, 'volume_size': 10}]
        models.assert_same_instance_flavors(test_instances)

        test_instances = [{'flavor_id': 1, 'volume_size': 10},
                          {'flavor_id': 1, 'volume_size': 10},
                          {'flavor_id': 1, 'volume_size': 10}]
        models.assert_same_instance_flavors(test_instances, required_flavor=1)

        test_instances = [{'flavor_id': 1, 'volume_size': 10},
                          {'flavor_id': 2, 'volume_size': 10},
                          {'flavor_id': 1, 'volume_size': 10}]
        self.assertRaises(exception.ClusterFlavorsNotEqual,
                          models.assert_same_instance_flavors,
                          test_instances)

        test_instances = [{'flavor_id': 1, 'volume_size': 10},
                          {'flavor_id': 1, 'volume_size': 10},
                          {'flavor_id': 1, 'volume_size': 10}]
        self.assertRaises(exception.ClusterFlavorsNotEqual,
                          models.assert_same_instance_flavors,
                          test_instances, required_flavor=2)

    @patch.multiple(models, assert_same_instance_flavors=DEFAULT,
                    assert_same_instance_volumes=DEFAULT)
    def test_assert_homogeneous_cluster(self, assert_same_instance_flavors,
                                        assert_same_instance_volumes):
        test_instances = [{'flavor_id': 1, 'volume_size': 10},
                          {'flavor_id': 1, 'volume_size': 10},
                          {'flavor_id': 1, 'volume_size': 10}]
        required_flavor = Mock()
        required_volume_size = Mock()
        models.assert_homogeneous_cluster(
            test_instances, required_flavor=required_flavor,
            required_volume_size=required_volume_size)
        assert_same_instance_flavors.assert_called_once_with(
            test_instances, required_flavor=required_flavor)
        assert_same_instance_volumes.assert_called_once_with(
            test_instances, required_size=required_volume_size)

    @patch.object(clients, 'create_neutron_client', return_value=MagicMock())
    def test_validate_instance_nics(self, create_neutron_cli_mock):

        test_instances = [
            {'volume_size': 1, 'flavor_id': '1234',
             'nics': [{"network_id": "surprise"}]},
            {'volume_size': 1, 'flavor_id': '1234',
             'nics': [{"network_id": "foo-bar"}]},
            {'volume_size': 1, 'flavor_id': '1234',
             'nics': [{"network_id": "foo-bar"}]}]

        self.assertRaises(exception.ClusterNetworksNotEqual,
                          models.validate_instance_nics,
                          Mock(),
                          test_instances)

        test_instances = [
            {'volume_size': 1, 'flavor_id': '1234',
             'nics': [{"network_id": "foo-bar"}]},
            {'volume_size': 1, 'flavor_id': '1234',
             'nics': [{"network_id": "foo-bar"}]},
            {'volume_size': 1, 'flavor_id': '1234',
             'nics': [{"network_id": "foo-bar"}]}]

        create_neutron_cli_mock.return_value.find_resource = Mock(
            side_effect=neutron_exceptions.NotFound(
                "Nic id not found %s" % id))

        self.assertRaises(exception.NetworkNotFound,
                          models.validate_instance_nics,
                          Mock(),
                          test_instances)



# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from urllib import parse as urllib_parse

from proboscis import SkipTest

from trove.common import exception
from trove.common.utils import poll_until
from trove.tests.scenario import runners
from trove.tests.scenario.runners.test_runners import SkipKnownBug
from trove.tests.scenario.runners.test_runners import TestRunner
from troveclient.compat import exceptions


class UserActionsRunner(TestRunner):

    # TODO(pmalik): I believe the 202 (Accepted) should be replaced by
    # 200 (OK) as the actions are generally very fast and their results
    # available immediately upon execution of the request. This would
    # likely require replacing GA casts with calls which I believe are
    # more appropriate anyways.

    def __init__(self):
        super(UserActionsRunner, self).__init__()
        self.user_defs = []
        self.renamed_user_orig_def = None

    @property
    def first_user_def(self):
        if self.user_defs:
            # Try to use the first user with databases if any.
            for user_def in self.user_defs:
                if 'databases' in user_def and user_def['databases']:
                    return user_def
            return self.user_defs[0]
        raise SkipTest("No valid user definitions provided.")

    @property
    def non_existing_user_def(self):
        user_def = self.test_helper.get_non_existing_user_definition()
        if user_def:
            return user_def
        raise SkipTest("No valid user definitions provided.")

    def run_users_create(self, expected_http_code=202):
        users = self.test_helper.get_valid_user_definitions()
        if users:
            self.user_defs = self.assert_users_create(
                self.instance_info.id, users, expected_http_code)
        else:
            raise SkipTest("No valid user definitions provided.")

    def assert_users_create(self, instance_id, serial_users_def,
                            expected_http_code):
        client = self.auth_client
        client.users.create(instance_id, serial_users_def)
        self.assert_client_code(client, expected_http_code)
        self.wait_for_user_create(client, instance_id, serial_users_def)
        return serial_users_def

    def run_user_show(self, expected_http_code=200):
        for user_def in self.user_defs:
            self.assert_user_show(
                self.instance_info.id, user_def, expected_http_code)

    def assert_user_show(self, instance_id, expected_user_def,
                         expected_http_code):
        user_name = expected_user_def['name']
        user_host = expected_user_def.get('host')

        client = self.auth_client
        queried_user = client.users.get(
            instance_id, user_name, user_host)
        self.assert_client_code(client, expected_http_code)
        self._assert_user_matches(queried_user, expected_user_def)

    def _assert_user_matches(self, user, expected_user_def):
        user_name = expected_user_def['name']
        self.assert_equal(expected_user_def['name'], user.name,
                          "Mismatch of names for user: %s" % user_name)
        self.assert_list_elements_equal(
            expected_user_def['databases'], user.databases,
            "Mismatch of databases for user: %s" % user_name)

    def run_users_list(self, expected_http_code=200):
        self.assert_users_list(
            self.instance_info.id, self.user_defs, expected_http_code)

    def assert_users_list(self, instance_id, expected_user_defs,
                          expected_http_code, limit=2):
        client = self.auth_client
        full_list = client.users.list(instance_id)
        self.assert_client_code(client, expected_http_code)
        listed_users = {user.name: user for user in full_list}
        self.assert_is_none(full_list.next,
                            "Unexpected pagination in the list.")

        for user_def in expected_user_defs:
            user_name = user_def['name']
            self.assert_true(
                user_name in listed_users,
                "User not included in the 'user-list' output: %s" %
                user_name)
            self._assert_user_matches(listed_users[user_name], user_def)

        # Check that the system (ignored) users are not included in the output.
        system_users = self.get_system_users()
        self.assert_false(
            any(name in listed_users for name in system_users),
            "System users should not be included in the 'user-list' output.")

        # Test list pagination.
        list_page = client.users.list(instance_id, limit=limit)
        self.assert_client_code(client, expected_http_code)

        self.assert_true(len(list_page) <= limit)
        if len(full_list) > limit:
            self.assert_is_not_none(list_page.next, "List page is missing.")
        else:
            self.assert_is_none(list_page.next, "An extra page in the list.")
        marker = list_page.next

        self.assert_pagination_match(list_page, full_list, 0, limit)
        if marker:
            last_user = list_page[-1]
            expected_marker = self.as_pagination_marker(last_user)
            self.assert_equal(expected_marker, marker,
                              "Pagination marker should be the last element "
                              "in the page.")
            list_page = client.users.list(instance_id, marker=marker)
            self.assert_client_code(client, expected_http_code)
            self.assert_pagination_match(
                list_page, full_list, limit, len(full_list))

    def as_pagination_marker(self, user):
        return urllib_parse.quote(user.name)

    def run_user_access_show(self, expected_http_code=200):
        for user_def in self.user_defs:
            self.assert_user_access_show(
                self.instance_info.id, user_def, expected_http_code)

    def assert_user_access_show(self, instance_id, user_def,
                                expected_http_code):
        user_name, user_host = self._get_user_name_host_pair(user_def)
        client = self.auth_client
        user_dbs = client.users.list_access(
            instance_id, user_name, hostname=user_host)
        self.assert_client_code(client, expected_http_code)

        expected_dbs = {db_def['name'] for db_def in user_def['databases']}
        listed_dbs = [db.name for db in user_dbs]

        self.assert_equal(len(expected_dbs), len(listed_dbs),
                          "Unexpected number of databases on the user access "
                          "list.")

        for database in expected_dbs:
            self.assert_true(
                database in listed_dbs,
                "Database not found in the user access list: %s" % database)

    def run_user_access_revoke(self, expected_http_code=202):
        self._apply_on_all_databases(
            self.instance_info.id, self.assert_user_access_revoke,
            expected_http_code)

    def _apply_on_all_databases(self, instance_id, action, expected_http_code):
        if any(user_def['databases'] for user_def in self.user_defs):
            for user_def in self.user_defs:
                user_name, user_host = self._get_user_name_host_pair(user_def)
                db_defs = user_def['databases']
                for db_def in db_defs:
                    db_name = db_def['name']
                    action(instance_id, user_name, user_host,
                           db_name, expected_http_code)
        else:
            raise SkipTest("No user databases defined.")

    def assert_user_access_revoke(self, instance_id, user_name, user_host,
                                  database, expected_http_code):
        client = self.auth_client
        client.users.revoke(
            instance_id, user_name, database, hostname=user_host)
        self.assert_client_code(client, expected_http_code)
        user_dbs = client.users.list_access(
            instance_id, user_name, hostname=user_host)
        self.assert_false(any(db.name == database for db in user_dbs),
                          "Database should no longer be included in the user "
                          "access list after revoke: %s" % database)

    def run_user_access_grant(self, expected_http_code=202):
        self._apply_on_all_databases(
            self.instance_info.id, self.assert_user_access_grant,
            expected_http_code)

    def assert_user_access_grant(self, instance_id, user_name, user_host,
                                 database, expected_http_code):
        client = self.auth_client
        client.users.grant(
            instance_id, user_name, [database], hostname=user_host)
        self.assert_client_code(client, expected_http_code)
        user_dbs = client.users.list_access(
            instance_id, user_name, hostname=user_host)
        self.assert_true(any(db.name == database for db in user_dbs),
                         "Database should be included in the user "
                         "access list after granting access: %s" % database)

    def run_user_create_with_no_attributes(
            self, expected_exception=exceptions.BadRequest,
            expected_http_code=400):
        self.assert_users_create_failure(
            self.instance_info.id, {}, expected_exception, expected_http_code)

    def run_user_create_with_blank_name(
            self, expected_exception=exceptions.BadRequest,
            expected_http_code=400):
        # Test with missing user name attribute.
        no_name_usr_def = self.copy_dict(self.non_existing_user_def,
                                         ignored_keys=['name'])
        self.assert_users_create_failure(
            self.instance_info.id, no_name_usr_def,
            expected_exception, expected_http_code)

        # Test with empty user name attribute.
        blank_name_usr_def = self.copy_dict(self.non_existing_user_def)
        blank_name_usr_def.update({'name': ''})
        self.assert_users_create_failure(
            self.instance_info.id, blank_name_usr_def,
            expected_exception, expected_http_code)

    def run_user_create_with_blank_password(
            self, expected_exception=exceptions.BadRequest,
            expected_http_code=400):
        # Test with missing password attribute.
        no_pass_usr_def = self.copy_dict(self.non_existing_user_def,
                                         ignored_keys=['password'])
        self.assert_users_create_failure(
            self.instance_info.id, no_pass_usr_def,
            expected_exception, expected_http_code)

        # Test with missing databases attribute.
        no_db_usr_def = self.copy_dict(self.non_existing_user_def,
                                       ignored_keys=['databases'])
        self.assert_users_create_failure(
            self.instance_info.id, no_db_usr_def,
            expected_exception, expected_http_code)

    def run_existing_user_create(
            self, expected_exception=exceptions.BadRequest,
            expected_http_code=400):
        self.assert_users_create_failure(
            self.instance_info.id, self.first_user_def,
            expected_exception, expected_http_code)

    def run_system_user_create(
            self, expected_exception=exceptions.BadRequest,
            expected_http_code=400):
        # TODO(pmalik): Actions on system users and databases should probably
        # return Forbidden 403 instead. The current error messages are
        # confusing (talking about a malformed request).
        system_users = self.get_system_users()
        if system_users:
            user_defs = [{'name': name, 'password': 'password1',
                          'databases': []} for name in system_users]
            self.assert_users_create_failure(
                self.instance_info.id, user_defs,
                expected_exception, expected_http_code)

    def assert_users_create_failure(
            self, instance_id, serial_users_def,
            expected_exception, expected_http_code):
        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.users.create, instance_id, serial_users_def)

    def run_user_update_with_blank_name(
            self, expected_exception=exceptions.BadRequest,
            expected_http_code=400):
        self.assert_user_attribute_update_failure(
            self.instance_info.id, self.first_user_def, {'name': ''},
            expected_exception, expected_http_code)

    def run_user_update_with_existing_name(
            self, expected_exception=exceptions.BadRequest,
            expected_http_code=400):
        self.assert_user_attribute_update_failure(
            self.instance_info.id, self.first_user_def,
            {'name': self.first_user_def['name']},
            expected_exception, expected_http_code)

    def assert_user_attribute_update_failure(
            self, instance_id, user_def, update_attribites,
            expected_exception, expected_http_code):
        user_name, user_host = self._get_user_name_host_pair(user_def)

        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.users.update_attributes, instance_id,
            user_name, update_attribites, user_host)

    def _get_user_name_host_pair(self, user_def):
        return user_def['name'], user_def.get('host')

    def run_system_user_attribute_update(
            self, expected_exception=exceptions.BadRequest,
            expected_http_code=400):
        # TODO(pmalik): Actions on system users and databases should probably
        # return Forbidden 403 instead. The current error messages are
        # confusing (talking about a malformed request).
        system_users = self.get_system_users()
        if system_users:
            for name in system_users:
                user_def = {'name': name, 'password': 'password2'}
                self.assert_user_attribute_update_failure(
                    self.instance_info.id, user_def, user_def,
                    expected_exception, expected_http_code)

    def run_user_attribute_update(self, expected_http_code=202):
        updated_def = self.first_user_def
        # Update the name by appending a random string to it.
        updated_name = ''.join([updated_def['name'], 'upd'])
        update_attribites = {'name': updated_name,
                             'password': 'password2'}
        self.assert_user_attribute_update(
            self.instance_info.id, updated_def,
            update_attribites, expected_http_code)

    def assert_user_attribute_update(self, instance_id, user_def,
                                     update_attribites, expected_http_code):
        user_name, user_host = self._get_user_name_host_pair(user_def)

        client = self.auth_client
        client.users.update_attributes(
            instance_id, user_name, update_attribites, user_host)
        self.assert_client_code(client, expected_http_code)

        # Update the stored definitions with the new value.
        expected_def = None
        for user_def in self.user_defs:
            if user_def['name'] == user_name:
                self.renamed_user_orig_def = dict(user_def)
                user_def.update(update_attribites)
                expected_def = user_def

        self.wait_for_user_create(client, instance_id, self.user_defs)

        # Verify using 'user-show' and 'user-list'.
        self.assert_user_show(instance_id, expected_def, 200)
        self.assert_users_list(instance_id, self.user_defs, 200)

    def run_user_recreate_with_no_access(self, expected_http_code=202):
        if (self.renamed_user_orig_def and
                self.renamed_user_orig_def['databases']):
            self.assert_user_recreate_with_no_access(
                self.instance_info.id, self.renamed_user_orig_def,
                expected_http_code)
        else:
            raise SkipTest("No renamed users with databases.")

    def assert_user_recreate_with_no_access(self, instance_id, original_def,
                                            expected_http_code=202):
        # Recreate a previously renamed user without assigning any access
        # rights to it.
        recreated_user_def = dict(original_def)
        recreated_user_def.update({'databases': []})
        user_def = self.assert_users_create(
            instance_id, [recreated_user_def], expected_http_code)

        # Append the new user to defs for cleanup.
        self.user_defs.extend(user_def)

        # Assert empty user access.
        self.assert_user_access_show(instance_id, recreated_user_def, 200)

    def run_user_delete(self, expected_http_code=202):
        for user_def in self.user_defs:
            self.assert_user_delete(
                self.instance_info.id, user_def, expected_http_code)

    def assert_user_delete(self, instance_id, user_def, expected_http_code):
        user_name, user_host = self._get_user_name_host_pair(user_def)

        client = self.auth_client
        client.users.delete(instance_id, user_name, user_host)
        self.assert_client_code(client, expected_http_code)
        self._wait_for_user_delete(client, instance_id, user_name)

    def _wait_for_user_delete(self, client, instance_id, deleted_user_name):
        self.report.log("Waiting for deleted user to disappear from the "
                        "listing: %s" % deleted_user_name)

        def _db_is_gone():
            all_users = self.get_user_names(client, instance_id)
            return deleted_user_name not in all_users

        try:
            poll_until(_db_is_gone, time_out=self.GUEST_CAST_WAIT_TIMEOUT_SEC)
            self.report.log("User is now gone from the instance.")
        except exception.PollTimeOut:
            self.fail("User still listed after the poll timeout: %ds" %
                      self.GUEST_CAST_WAIT_TIMEOUT_SEC)

    def run_nonexisting_user_show(
            self, expected_exception=exceptions.NotFound,
            expected_http_code=404):
        self.assert_user_show_failure(
            self.instance_info.id,
            {'name': self.non_existing_user_def['name']},
            expected_exception, expected_http_code)

    def assert_user_show_failure(self, instance_id, user_def,
                                 expected_exception, expected_http_code):
        user_name, user_host = self._get_user_name_host_pair(user_def)

        client = self.auth_client
        self.assert_raises(
            expected_exception, expected_http_code,
            client, client.users.get, instance_id, user_name, user_host)

    def run_system_user_show(
            self, expected_exception=exceptions.BadRequest,
            expected_http_code=400):
        # TODO(pmalik): Actions on system users and databases should probably
        # return Forbidden 403 instead. The current error messages are
        # confusing (talking about a malformed request).
        system_users = self.get_system_users()
        if system_users:
            for name in system_users:
                self.assert_user_show_failure(
                    self.instance_info.id, {'name': name},
                    expected_exception, expected_http_code)

    def run_nonexisting_user_update(self, expected_http_code=404):
        # Test valid update on a non-existing user.
        update_def = {'name': self.non_existing_user_def['name']}
        self.assert_user_attribute_update_failure(
            self.instance_info.id, update_def, update_def,
            exceptions.NotFound, expected_http_code)

    def run_nonexisting_user_delete(
            self, expected_exception=exceptions.NotFound,
            expected_http_code=404):
        self.assert_user_delete_failure(
            self.instance_info.id,
            {'name': self.non_existing_user_def['name']},
            expected_exception, expected_http_code)

    def assert_user_delete_failure(
            self, instance_id, user_def,
            expected_exception, expected_http_code):
        user_name, user_host = self._get_user_name_host_pair(user_def)

        client = self.auth_client
        self.assert_raises(expected_exception, expected_http_code,
                           client, client.users.delete,
                           instance_id, user_name, user_host)

    def run_system_user_delete(
            self, expected_exception=exceptions.BadRequest,
            expected_http_code=400):
        # TODO(pmalik): Actions on system users and databases should probably
        # return Forbidden 403 instead. The current error messages are
        # confusing (talking about a malformed request).
        system_users = self.get_system_users()
        if system_users:
            for name in system_users:
                self.assert_user_delete_failure(
                    self.instance_info.id, {'name': name},
                    expected_exception, expected_http_code)

    def get_system_users(self):
        return self.get_datastore_config_property('ignore_users')


class MysqlUserActionsRunner(UserActionsRunner):

    def as_pagination_marker(self, user):
        return urllib_parse.quote('%s@%s' % (user.name, user.host))


class MariadbUserActionsRunner(MysqlUserActionsRunner):

    def __init__(self):
        super(MariadbUserActionsRunner, self).__init__()


class PerconaUserActionsRunner(MysqlUserActionsRunner):

    def __init__(self):
        super(PerconaUserActionsRunner, self).__init__()


class PxcUserActionsRunner(MysqlUserActionsRunner):

    def __init__(self):
        super(PxcUserActionsRunner, self).__init__()


class PostgresqlUserActionsRunner(UserActionsRunner):

    def run_user_update_with_existing_name(self):
        raise SkipKnownBug(runners.BUG_WRONG_API_VALIDATION)

    def run_system_user_show(self):
        raise SkipKnownBug(runners.BUG_WRONG_API_VALIDATION)

    def run_system_user_attribute_update(self):
        raise SkipKnownBug(runners.BUG_WRONG_API_VALIDATION)

    def run_system_user_delete(self):
        raise SkipKnownBug(runners.BUG_WRONG_API_VALIDATION)



# Copyright 2013 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from xml.dom import minidom
from xml.parsers import expat
from xml import sax
from xml.sax import expatreader

from trove.common.i18n import _


class ProtectedExpatParser(expatreader.ExpatParser):
    """An expat parser which disables DTD's and entities by default."""

    def __init__(self, forbid_dtd=True, forbid_entities=True,
                 *args, **kwargs):
        # Python 2.x old style class
        expatreader.ExpatParser.__init__(self, *args, **kwargs)
        self.forbid_dtd = forbid_dtd
        self.forbid_entities = forbid_entities

    def start_doctype_decl(self, name, sysid, pubid, has_internal_subset):
        raise ValueError(_("Inline DTD forbidden"))

    def entity_decl(self, entityName, is_parameter_entity, value, base,
                    systemId, publicId, notationName):
        raise ValueError(_("<!ENTITY> entity declaration forbidden"))

    def unparsed_entity_decl(self, name, base, sysid, pubid, notation_name):
        # expat 1.2
        raise ValueError(_("<!ENTITY> unparsed entity forbidden"))

    def external_entity_ref(self, context, base, systemId, publicId):
        raise ValueError(_("<!ENTITY> external entity forbidden"))

    def notation_decl(self, name, base, sysid, pubid):
        raise ValueError(_("<!ENTITY> notation forbidden"))

    def reset(self):
        expatreader.ExpatParser.reset(self)
        if self.forbid_dtd:
            self._parser.StartDoctypeDeclHandler = self.start_doctype_decl
            self._parser.EndDoctypeDeclHandler = None
        if self.forbid_entities:
            self._parser.EntityDeclHandler = self.entity_decl
            self._parser.UnparsedEntityDeclHandler = self.unparsed_entity_decl
            self._parser.ExternalEntityRefHandler = self.external_entity_ref
            self._parser.NotationDeclHandler = self.notation_decl
            try:
                self._parser.SkippedEntityHandler = None
            except AttributeError:
                # some pyexpat versions do not support SkippedEntity
                pass


def safe_minidom_parse_string(xml_string):
    """Parse an XML string using minidom safely.

    """
    try:
        return minidom.parseString(xml_string, parser=ProtectedExpatParser())
    except sax.SAXParseException:
        raise expat.ExpatError()



# Copyright 2014 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utility methods for working with WSGI servers."""

import eventlet
eventlet.patcher.monkey_patch(all=False, socket=True)

import datetime
import errno
import socket
import sys
import time

import eventlet.wsgi
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_service import service
from oslo_service import sslutils
import routes
import routes.middleware
import webob.dec
import webob.exc
from xml.dom import minidom
from xml.parsers import expat

from trove.common import base_exception
from trove.common.i18n import _
from trove.common.utils import req_to_text
from trove.common import xmlutils

socket_opts = [
    cfg.IntOpt('backlog',
               default=4096,
               help="Number of backlog requests to configure the socket with"),
    cfg.IntOpt('tcp_keepidle',
               default=600,
               help="Sets the value of TCP_KEEPIDLE in seconds for each "
                    "server socket. Not supported on OS X."),
]

CONF = cfg.CONF
CONF.register_opts(socket_opts)

LOG = logging.getLogger(__name__)


def run_server(application, port, **kwargs):
    """Run a WSGI server with the given application."""
    sock = eventlet.listen(('0.0.0.0', port))
    eventlet.wsgi.server(sock, application, **kwargs)


class Service(service.Service):
    """
    Provides a Service API for wsgi servers.

    This gives us the ability to launch wsgi servers with the
    Launcher classes in oslo_service.service.py.
    """

    def __init__(self, application, port,
                 host='0.0.0.0', backlog=4096, threads=1000):
        self.application = application
        self._port = port
        self._host = host
        self._backlog = backlog if backlog else CONF.backlog
        self._socket = self._get_socket(host, port, self._backlog)
        super(Service, self).__init__(threads)

    def _get_socket(self, host, port, backlog):
        # TODO(dims): eventlet's green dns/socket module does not actually
        # support IPv6 in getaddrinfo(). We need to get around this in the
        # future or monitor upstream for a fix
        info = socket.getaddrinfo(host,
                                  port,
                                  socket.AF_UNSPEC,
                                  socket.SOCK_STREAM)[0]
        family = info[0]
        bind_addr = info[-1]

        sock = None
        retry_until = time.time() + 30
        while not sock and time.time() < retry_until:
            try:
                sock = eventlet.listen(bind_addr,
                                       backlog=backlog,
                                       family=family)
                if sslutils.is_enabled(CONF):
                    sock = sslutils.wrap(CONF, sock)

            except socket.error as err:
                if err.args[0] != errno.EADDRINUSE:
                    raise
                eventlet.sleep(0.1)
        if not sock:
            raise RuntimeError(_("Could not bind to %(host)s:%(port)s "
                               "after trying for 30 seconds") %
                               {'host': host, 'port': port})
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # sockets can hang around forever without keepalive
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)

        # This option isn't available in the OS X version of eventlet
        if hasattr(socket, 'TCP_KEEPIDLE'):
            sock.setsockopt(socket.IPPROTO_TCP,
                            socket.TCP_KEEPIDLE,
                            CONF.tcp_keepidle)

        return sock

    def start(self):
        """Start serving this service using the provided server instance.

        :returns: None

        """
        super(Service, self).start()
        self.tg.add_thread(self._run, self.application, self._socket)

    @property
    def backlog(self):
        return self._backlog

    @property
    def host(self):
        return self._socket.getsockname()[0] if self._socket else self._host

    @property
    def port(self):
        return self._socket.getsockname()[1] if self._socket else self._port

    def stop(self):
        """Stop serving this API.

        :returns: None

        """
        super(Service, self).stop()

    def _run(self, application, socket):
        """Start a WSGI server in a new green thread."""
        logger = logging.getLogger('eventlet.wsgi')
        eventlet.wsgi.server(socket,
                             application,
                             custom_pool=self.tg.pool,
                             log=logger)


class Middleware(object):
    """
    Base WSGI middleware wrapper. These classes require an application to be
    initialized that will be called next.  By default the middleware will
    simply call its wrapped app, or you can override __call__ to customize its
    behavior.
    """

    def __init__(self, application):
        self.application = application

    def process_request(self, req):
        """
        Called on each request.

        If this returns None, the next application down the stack will be
        executed. If it returns a response then that response will be returned
        and execution will stop here.
        """
        return None

    def process_response(self, response):
        """Do whatever you'd like to the response."""
        return response

    @webob.dec.wsgify
    def __call__(self, req):
        response = self.process_request(req)
        if response:
            return response
        response = req.get_response(self.application)
        return self.process_response(response)


class Debug(Middleware):
    """
    Helper class that can be inserted into any WSGI application chain
    to get information about the request and response.
    """

    @webob.dec.wsgify
    def __call__(self, req):
        print(("*" * 40) + " REQUEST ENVIRON")
        for key, value in req.environ.items():
            print(key, "=", value)
        print()
        resp = req.get_response(self.application)

        print(("*" * 40) + " RESPONSE HEADERS")
        for (key, value) in resp.headers.items():
            print(key, "=", value)
        print()

        resp.app_iter = self.print_generator(resp.app_iter)

        return resp

    @staticmethod
    def print_generator(app_iter):
        """
        Iterator that prints the contents of a wrapper string iterator
        when iterated.
        """
        print(("*" * 40) + " BODY")
        for part in app_iter:
            sys.stdout.write(part)
            sys.stdout.flush()
            yield part
        print()


class Router(object):

    """
    WSGI middleware that maps incoming requests to WSGI apps.
    """

    def __init__(self, mapper):
        """
        Create a router for the given routes.Mapper.

        Each route in `mapper` must specify a 'controller', which is a
        WSGI app to call.  You'll probably want to specify an 'action' as
        well and have your controller be a wsgi.Controller, who will route
        the request to the action method.

        Examples:
          mapper = routes.Mapper()
          sc = ServerController()

          # Explicit mapping of one route to a controller+action
          mapper.connect(None, "/svrlist", controller=sc, action="list")

          # Actions are all implicitly defined
          mapper.resource("server", "servers", controller=sc)

          # Pointing to an arbitrary WSGI app.  You can specify the
          # {path_info:.*} parameter so the target app can be handed just that
          # section of the URL.
          mapper.connect(None, "/v1.0/{path_info:.*}", controller=BlogApp())
        """
        self.map = mapper
        self._router = routes.middleware.RoutesMiddleware(self._dispatch,
                                                          self.map)

    @webob.dec.wsgify
    def __call__(self, req):
        """
        Route the incoming request to a controller based on self.map.
        If no match, return a 404.
        """
        return self._router

    @staticmethod
    @webob.dec.wsgify
    def _dispatch(req):
        """
        Called by self._router after matching the incoming request to a route
        and putting the information into req.environ.  Either returns 404
        or the routed WSGI app's response.
        """
        match = req.environ['wsgiorg.routing_args'][1]
        if not match:
            return webob.exc.HTTPNotFound()
        app = match['controller']
        return app


class Request(webob.Request):
    """Add some OpenStack API-specific logic to the base webob.Request."""

    default_request_content_types = ('application/json', 'application/xml')
    default_accept_types = ('application/json', 'application/xml')
    default_accept_type = 'application/json'

    def best_match_content_type(self, supported_content_types=None):
        """Determine the requested response content-type.

        Based on the query extension then the Accept header.
        Defaults to default_accept_type if we don't find a preference

        """
        supported_content_types = (supported_content_types or
                                   self.default_accept_types)

        parts = self.path.rsplit('.', 1)
        if len(parts) > 1:
            ctype = 'application/{0}'.format(parts[1])
            if ctype in supported_content_types:
                return ctype

        bm = self.accept.best_match(supported_content_types)
        return bm or self.default_accept_type

    def get_content_type(self, allowed_content_types=None):
        """Determine content type of the request body.

        Does not do any body introspection, only checks header

        """
        if "Content-Type" not in self.headers:
            return None

        content_type = self.content_type
        allowed_content_types = (allowed_content_types or
                                 self.default_request_content_types)

        if content_type not in allowed_content_types:
            raise base_exception.InvalidContentType(content_type=content_type)
        return content_type

    __str__ = req_to_text


class Resource(object):
    """
    WSGI app that handles (de)serialization and controller dispatch.

    Reads routing information supplied by RoutesMiddleware and calls
    the requested action method upon its deserializer, controller,
    and serializer. Those three objects may implement any of the basic
    controller action methods (create, update, show, index, delete)
    along with any that may be specified in the api router. A 'default'
    method may also be implemented to be used in place of any
    non-implemented actions. Deserializer methods must accept a request
    argument and return a dictionary. Controller methods must accept a
    request argument. Additionally, they must also accept keyword
    arguments that represent the keys returned by the Deserializer. They
    may raise a webob.exc exception or return a dict, which will be
    serialized by requested content type.
    """
    def __init__(self, controller, deserializer=None, serializer=None):
        """
        :param controller: object that implement methods created by routes lib
        :param deserializer: object that supports webob request deserialization
                             through controller-like actions
        :param serializer: object that supports webob response serialization
                           through controller-like actions
        """
        self.controller = controller
        self.serializer = serializer or ResponseSerializer()
        self.deserializer = deserializer or RequestDeserializer()

    @webob.dec.wsgify(RequestClass=Request)
    def __call__(self, request):
        """WSGI method that controls (de)serialization and method dispatch."""

        try:
            action, action_args, accept = self.deserialize_request(request)
        except base_exception.InvalidContentType:
            msg = _("Unsupported Content-Type")
            return webob.exc.HTTPUnsupportedMediaType(explanation=msg)
        except base_exception.MalformedRequestBody:
            msg = _("Malformed request body")
            return webob.exc.HTTPBadRequest(explanation=msg)

        action_result = self.execute_action(action, request, **action_args)
        try:
            return self.serialize_response(action, action_result, accept)
        # return unserializable result (typically a webob exc)
        except Exception:
            return action_result

    def deserialize_request(self, request):
        return self.deserializer.deserialize(request)

    def serialize_response(self, action, action_result, accept):
        return self.serializer.serialize(action_result, accept, action)

    def execute_action(self, action, request, **action_args):
        return self.dispatch(self.controller, action, request, **action_args)

    def dispatch(self, obj, action, *args, **kwargs):
        """Find action-specific method on self and call it."""
        try:
            method = getattr(obj, action)
        except AttributeError:
            method = getattr(obj, 'default')

        return method(*args, **kwargs)

    def get_action_args(self, request_environment):
        """Parse dictionary created by routes library."""
        try:
            args = request_environment['wsgiorg.routing_args'][1].copy()
        except Exception:
            return {}

        try:
            del args['controller']
        except KeyError:
            pass

        try:
            del args['format']
        except KeyError:
            pass

        return args


class ActionDispatcher(object):
    """Maps method name to local methods through action name."""

    def dispatch(self, *args, **kwargs):
        """Find and call local method."""
        action = kwargs.pop('action', 'default')
        action_method = getattr(self, str(action), self.default)
        return action_method(*args, **kwargs)

    def default(self, data):
        raise NotImplementedError()


class DictSerializer(ActionDispatcher):
    """Default request body serialization."""

    def serialize(self, data, action='default'):
        return self.dispatch(data, action=action)

    def default(self, data):
        return ""


class JSONDictSerializer(DictSerializer):
    """Default JSON request body serialization."""

    def default(self, data):
        def sanitizer(obj):
            if isinstance(obj, datetime.datetime):
                _dtime = obj - datetime.timedelta(microseconds=obj.microsecond)
                return _dtime.isoformat()
            return obj
        return jsonutils.dump_as_bytes(data, default=sanitizer)


class XMLDictSerializer(DictSerializer):

    def __init__(self, metadata=None, xmlns=None):
        """
        :param metadata: information needed to deserialize xml into
                         a dictionary.
        :param xmlns: XML namespace to include with serialized xml
        """
        super(XMLDictSerializer, self).__init__()
        self.metadata = metadata or {}
        self.xmlns = xmlns

    def default(self, data):
        # We expect data to contain a single key which is the XML root.
        root_key = list(data.keys())[0]
        doc = minidom.Document()
        node = self._to_xml_node(doc, self.metadata, root_key, data[root_key])

        return self.to_xml_string(node)

    def to_xml_string(self, node, has_atom=False):
        self._add_xmlns(node, has_atom)
        return node.toprettyxml(indent='    ', encoding='UTF-8')

    # NOTE (ameade): the has_atom should be removed after all of the
    # xml serializers and view builders have been updated to the current
    # spec that required all responses include the xmlns:atom, the has_atom
    # flag is to prevent current tests from breaking
    def _add_xmlns(self, node, has_atom=False):
        if self.xmlns is not None:
            node.setAttribute('xmlns', self.xmlns)
        if has_atom:
            node.setAttribute('xmlns:atom', "http://www.w3.org/2005/Atom")

    def _to_xml_node(self, doc, metadata, nodename, data):
        """Recursive method to convert data members to XML nodes."""
        result = doc.createElement(nodename)

        # Set the xml namespace if one is specified
        # TODO(justinsb): We could also use prefixes on the keys
        xmlns = metadata.get('xmlns', None)
        if xmlns:
            result.setAttribute('xmlns', xmlns)

        # TODO(bcwaldon): accomplish this without a type-check
        if type(data) is list:
            collections = metadata.get('list_collections', {})
            if nodename in collections:
                metadata = collections[nodename]
                for item in data:
                    node = doc.createElement(metadata['item_name'])
                    node.setAttribute(metadata['item_key'], str(item))
                    result.appendChild(node)
                return result
            singular = metadata.get('plurals', {}).get(nodename, None)
            if singular is None:
                if nodename.endswith('s'):
                    singular = nodename[:-1]
                else:
                    singular = 'item'
            for item in data:
                node = self._to_xml_node(doc, metadata, singular, item)
                result.appendChild(node)
        # TODO(bcwaldon): accomplish this without a type-check
        elif type(data) is dict:
            collections = metadata.get('dict_collections', {})
            if nodename in collections:
                metadata = collections[nodename]
                for k, v in data.items():
                    node = doc.createElement(metadata['item_name'])
                    node.setAttribute(metadata['item_key'], str(k))
                    text = doc.createTextNode(str(v))
                    node.appendChild(text)
                    result.appendChild(node)
                return result
            attrs = metadata.get('attributes', {}).get(nodename, {})
            for k, v in data.items():
                if k in attrs:
                    result.setAttribute(k, str(v))
                else:
                    node = self._to_xml_node(doc, metadata, k, v)
                    result.appendChild(node)
        else:
            # Type is atom
            node = doc.createTextNode(str(data))
            result.appendChild(node)
        return result

    def _create_link_nodes(self, xml_doc, links):
        link_nodes = []
        for link in links:
            link_node = xml_doc.createElement('atom:link')
            link_node.setAttribute('rel', link['rel'])
            link_node.setAttribute('href', link['href'])
            if 'type' in link:
                link_node.setAttribute('type', link['type'])
            link_nodes.append(link_node)
        return link_nodes


class ResponseHeadersSerializer(ActionDispatcher):
    """Default response headers serialization."""

    def serialize(self, response, data, action):
        self.dispatch(response, data, action=action)

    def default(self, response, data):
        response.status_int = 200


class ResponseSerializer(object):
    """Encode the necessary pieces into a response object."""

    def __init__(self, body_serializers=None, headers_serializer=None):
        self.body_serializers = {
            'application/xml': XMLDictSerializer(),
            'application/json': JSONDictSerializer(),
        }
        self.body_serializers.update(body_serializers or {})

        self.headers_serializer = (headers_serializer or
                                   ResponseHeadersSerializer())

    def serialize(self, response_data, content_type, action='default'):
        """Serialize a dict into a string and wrap in a wsgi.Request object.

        :param response_data: dict produced by the Controller
        :param content_type: expected mimetype of serialized response body

        """
        response = webob.Response()
        self.serialize_headers(response, response_data, action)
        self.serialize_body(response, response_data, content_type, action)
        return response

    def serialize_headers(self, response, data, action):
        self.headers_serializer.serialize(response, data, action)

    def serialize_body(self, response, data, content_type, action):
        response.headers['Content-Type'] = content_type
        if data is not None:
            serializer = self.get_body_serializer(content_type)
            response.body = serializer.serialize(data, action)

    def get_body_serializer(self, content_type):
        try:
            return self.body_serializers[content_type]
        except (KeyError, TypeError):
            raise base_exception.InvalidContentType(content_type=content_type)


class RequestHeadersDeserializer(ActionDispatcher):
    """Default request headers deserializer"""

    def deserialize(self, request, action):
        return self.dispatch(request, action=action)

    def default(self, request):
        return {}


class RequestDeserializer(object):
    """Break up a Request object into more useful pieces."""

    def __init__(self, body_deserializers=None, headers_deserializer=None,
                 supported_content_types=None):

        self.supported_content_types = supported_content_types

        self.body_deserializers = {
            'application/xml': XMLDeserializer(),
            'application/json': JSONDeserializer(),
        }
        self.body_deserializers.update(body_deserializers or {})

        self.headers_deserializer = (headers_deserializer or
                                     RequestHeadersDeserializer())

    def deserialize(self, request):
        """Extract necessary pieces of the request.

        :param request: Request object
        :returns: tuple of (expected controller action name, dictionary of
                  keyword arguments to pass to the controller, the expected
                  content type of the response)

        """
        action_args = self.get_action_args(request.environ)
        action = action_args.pop('action', None)

        action_args.update(self.deserialize_headers(request, action))
        action_args.update(self.deserialize_body(request, action))

        accept = self.get_expected_content_type(request)

        return (action, action_args, accept)

    def deserialize_headers(self, request, action):
        return self.headers_deserializer.deserialize(request, action)

    def deserialize_body(self, request, action):
        if not len(request.body) > 0:
            LOG.debug("Empty body provided in request")
            return {}

        try:
            content_type = request.get_content_type()
        except base_exception.InvalidContentType:
            LOG.debug("Unrecognized Content-Type provided in request")
            raise

        if content_type is None:
            LOG.debug("No Content-Type provided in request")
            return {}

        try:
            deserializer = self.get_body_deserializer(content_type)
        except base_exception.InvalidContentType:
            LOG.debug("Unable to deserialize body as provided Content-Type")
            raise

        return deserializer.deserialize(request.body, action)

    def get_body_deserializer(self, content_type):
        try:
            return self.body_deserializers[content_type]
        except (KeyError, TypeError):
            raise base_exception.InvalidContentType(content_type=content_type)

    def get_expected_content_type(self, request):
        return request.best_match_content_type(self.supported_content_types)

    def get_action_args(self, request_environment):
        """Parse dictionary created by routes library."""
        try:
            args = request_environment['wsgiorg.routing_args'][1].copy()
        except Exception:
            return {}

        try:
            del args['controller']
        except KeyError:
            pass

        try:
            del args['format']
        except KeyError:
            pass

        return args


class TextDeserializer(ActionDispatcher):
    """Default request body deserialization."""

    def deserialize(self, datastring, action='default'):
        return self.dispatch(datastring, action=action)

    def default(self, datastring):
        return {}


class JSONDeserializer(TextDeserializer):

    def _from_json(self, datastring):
        try:
            return jsonutils.loads(datastring)
        except ValueError:
            msg = _("cannot understand JSON")
            raise base_exception.MalformedRequestBody(reason=msg)

    def default(self, datastring):
        return {'body': self._from_json(datastring)}


class XMLDeserializer(TextDeserializer):

    def __init__(self, metadata=None):
        """
        :param metadata: information needed to deserialize xml into
                         a dictionary.
        """
        super(XMLDeserializer, self).__init__()
        self.metadata = metadata or {}

    def _from_xml(self, datastring):
        plurals = set(self.metadata.get('plurals', {}))

        try:
            node = xmlutils.safe_minidom_parse_string(datastring).childNodes[0]
            return {node.nodeName: self._from_xml_node(node, plurals)}
        except expat.ExpatError:
            msg = _("cannot understand XML")
            raise base_exception.MalformedRequestBody(reason=msg)

    def _from_xml_node(self, node, listnames):
        """Convert a minidom node to a simple Python type.

        :param listnames: list of XML node names whose subnodes should
                          be considered list items.

        """

        if len(node.childNodes) == 1 and node.childNodes[0].nodeType == 3:
            return node.childNodes[0].nodeValue
        elif node.nodeName in listnames:
            return [self._from_xml_node(n, listnames) for n in node.childNodes]
        else:
            result = dict()
            for attr in node.attributes.keys():
                result[attr] = node.attributes[attr].nodeValue
            for child in node.childNodes:
                if child.nodeType != node.TEXT_NODE:
                    result[child.nodeName] = self._from_xml_node(child,
                                                                 listnames)
            return result

    def find_first_child_named(self, parent, name):
        """Search a nodes children for the first child with a given name."""
        for node in parent.childNodes:
            if node.nodeName == name:
                return node
        return None

    def find_children_named(self, parent, name):
        """Return all of a nodes children who have the given name."""
        for node in parent.childNodes:
            if node.nodeName == name:
                yield node

    def extract_text(self, node):
        """Get the text field contained by the given node."""
        if len(node.childNodes) == 1:
            child = node.childNodes[0]
            if child.nodeType == child.TEXT_NODE:
                return child.nodeValue
        return ""

    def default(self, datastring):
        return {'body': self._from_xml(datastring)}



# Copyright 2016 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from trove.common.rpc import secure_serializer as ssz
from trove.tests.unittests import trove_testtools


class TestSecureSerializer(trove_testtools.TestCase):

    def setUp(self):
        self.key = 'xuUyAKn5mDANoM5sRxQsb6HGiugWVD'
        self.data = '5rzFfaKU630rRxL1g3c80EHnHDf534'
        self.context = {'fld1': 3, 'fld2': 'abc'}
        super(TestSecureSerializer, self).setUp()

    def tearDown(self):
        super(TestSecureSerializer, self).tearDown()

    def test_sz_nokey_serialize_entity(self):
        sz = ssz.SecureSerializer(base=None, key=None)
        en = sz.serialize_entity(self.context, self.data)
        self.assertEqual(en, self.data)

    def test_sz_nokey_deserialize_entity(self):
        sz = ssz.SecureSerializer(base=None, key=None)
        en = sz.deserialize_entity(self.context, self.data)
        self.assertEqual(en, self.data)

    def test_sz_nokey_serialize_context(self):
        sz = ssz.SecureSerializer(base=None, key=None)
        en = sz.serialize_context(self.context)
        self.assertEqual(en, self.context)

    def test_sz_nokey_deserialize_context(self):
        sz = ssz.SecureSerializer(base=None, key=None)
        en = sz.deserialize_context(self.context)
        self.assertEqual(en, self.context)

    def test_sz_entity(self):
        sz = ssz.SecureSerializer(base=None, key=self.key)
        en = sz.serialize_entity(self.context, self.data)
        self.assertNotEqual(en, self.data)
        self.assertEqual(sz.deserialize_entity(self.context, en),
                         self.data)

    def test_sz_context(self):
        sz = ssz.SecureSerializer(base=None, key=self.key)
        sctxt = sz.serialize_context(self.context)
        self.assertNotEqual(sctxt, self.context)
        self.assertEqual(sz.deserialize_context(sctxt),
                         self.context)
